"""This module contains the classes for the plot widgets, based on pyqtgraph."""

import sys
import os

sys.path.append(os.path.dirname(__file__).split("nomad_camels")[0])
import numpy as np
import threading
from collections import deque

from PySide6.QtWidgets import (
    QWidget,
    QGridLayout,
    QPushButton,
    QLabel,
    QLineEdit,
    QMenuBar,
    QGraphicsSceneMouseEvent,
    QTableWidgetItem,
    QComboBox,
    QColorDialog,
    QApplication,
)
from PySide6.QtCore import Signal, QObject, QEvent, Qt, QCoreApplication
from PySide6.QtGui import QIcon, QColor
import PySide6
import pyqtgraph as pg
from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent

import lmfit
from importlib import resources
from nomad_camels import graphics

from bluesky.callbacks.core import get_obj_fields, CallbackBase

from nomad_camels.gui.plot_options import Ui_Plot_Options
from nomad_camels.utility.fit_variable_renaming import replace_name
from nomad_camels.bluesky_handling.evaluation_helper import Evaluator
from nomad_camels.main_classes.plot_widget import LiveFit_Eva
from nomad_camels.utility.plot_placement import place_widget

# recognized by pyqtgraph: r, g, b, c, m, y, k, w
dark_mode_colors = ["w", "r", (0, 100, 255), "g", "c", "m", "y", "k"]

# these are the colors used by matplotlib, they are used as default colors in light mode
matplotlib_default_colors = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]

# these are the symbols recognized by pyqtgraph
symbols = {
    "circle": "o",
    "square": "s",
    "triangle": "t",
    "diamond": "d",
    "plus": "+",
    "upwards triangle": "t1",
    "right triangle": "t2",
    "left triangle": "t3",
    "pentagon": "p",
    "hexagon": "h",
    "star": "star",
    "cross": "x",
    "arrow_up": "arrow_up",
    "arrow_right": "arrow_right",
    "arrow_down": "arrow_down",
    "arrow_left": "arrow_left",
    "crosshair": "crosshair",
    "none": None,
}

# these are the linestyles recognized by pyqtgraph
linestyles = {
    "solid": Qt.PenStyle.SolidLine,
    "dashed": Qt.PenStyle.DashLine,
    "dash-dot": Qt.PenStyle.DashDotLine,
    "dash-dot-dot": Qt.PenStyle.DashDotDotLine,
    "dotted": Qt.PenStyle.DotLine,
    "none": Qt.PenStyle.NoPen,
}

dark_mode = False
pg.setConfigOptions(background="w", foreground="k")
colors = matplotlib_default_colors


def activate_dark_mode():
    """Changes the plot-style to dark-mode, by changing the config options of pyqtgraph."""
    global dark_mode, colors
    dark_mode = True
    pg.setConfigOptions(background="k", foreground="w")
    colors = dark_mode_colors


class ListDeque_skip:
    def __init__(self, iterable=None, maxlen=None, skip_n_points=0):
        if iterable is None:
            iterable = []
        self.maxlen = maxlen
        self.skip_n_points = skip_n_points
        self.counter_value = 0
        if (
            maxlen is None
            or maxlen == np.inf
            or (isinstance(maxlen, str) and maxlen.lower() in ["none", "inf", "np.inf"])
        ):
            self.data = list(iterable)
        else:
            self.data = deque(iterable, maxlen=maxlen)

    def append(self, item):
        self.counter_value += 1
        if self.skip_n_points <= 0 or self.counter_value % self.skip_n_points == 0:
            self.data.append(item)

    def extend(self, iterable):
        self.data.extend(iterable)

    def __getitem__(self, index):
        return self.data[index]

    def __setitem__(self, index, value):
        self.data[index] = value

    def __delitem__(self, index):
        del self.data[index]

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __repr__(self):
        return repr(self.data)

    def clear(self):
        self.data.clear()

    def pop(self):
        return self.data.pop()

    def popleft(self):
        if isinstance(self.data, deque):
            return self.data.popleft()
        else:
            raise AttributeError("'list' object has no attribute 'popleft'")

    def appendleft(self, item):
        if isinstance(self.data, deque):
            self.data.appendleft(item)
        else:
            raise AttributeError("'list' object has no attribute 'appendleft'")

    def extendleft(self, iterable):
        if isinstance(self.data, deque):
            self.data.extendleft(iterable)
        else:
            raise AttributeError("'list' object has no attribute 'extendleft'")

    def rotate(self, n=1):
        if isinstance(self.data, deque):
            self.data.rotate(n)
        else:
            raise AttributeError("'list' object has no attribute 'rotate'")

    def change_maxlen(self, maxlen):
        self.maxlen = maxlen
        if maxlen is None or maxlen == np.inf:
            self.data = list(self.data)
        else:
            self.data = deque(self.data, maxlen=maxlen)

    def change_skip_n_points(self, skip_n_points):
        self.skip_n_points = skip_n_points
        self.counter_value = 0


class PlotWidget(QWidget):
    """Class for creating a plot widget.

    Parameters
    ----------
    x_name : str
        The name of the x-axis variable
    y_names : Union[str, Tuple[str]]
        The name(s) of the y-axis variable(s)
    legend_keys : List[str]
        deprecated
    xlim : Tuple[float, float]
        deprecated
    ylim : Tuple[float, float]
        deprecated
    epoch : {'run', 'unix'}
        If 'run' t=0 is the time recorded in the RunStart document. If 'unix',
        t=0 is 1 Jan 1970 ("the UNIX epoch"). Default is 'run'.
    parent : QWidget, optional
        The parent widget, by default None
    namespace : Mapping[str, Any], optional
        The namespace to use for the `Evaluator`, by default None
    ylabel : str, optional
        The y-axis label, if empty the first y_name is used, by default ''
    xlabel : str, optional
        The x-axis label, if empty the x_name is used, by default ''
    title : str, optional
        The title of the plot, by default ''
    stream_name : str
        The name of the bluesky stream to be used for the plot. If multi_stream is True, streams including this name are used. Default is 'primary'
    fits : List[Dict[str, Union[str, bool, List[str], Tuple[float, float], Dict[str, Union[str, float]]]]], optional
        The fits for the plot, by default None
    do_plot : bool
        deprecated
    multi_stream : bool, optional
        Whether to use multiple streams, see stream name, by default False
    y_axes : Dict[str, int], optional
        The y-axis to use for each y_name, the ints should be 1 or 2, if 2, the respective y-value is plotted on the right axis, by default None
    logX : bool
        Whether to use a logarithmic x-axis, by default False
    logY : bool
        Whether to use a logarithmic y-axis, by default False
    logY2 : bool
        Whether to use a logarithmic y-axis for the right axis, by default False
    maxlen : int
        The maximum number of data points to show, by default np.inf
    use_bluesky : bool
    **kwargs : Any, optional
        Additional keyword arguments to pass to `MultiLivePlot`
    """

    closing = Signal()
    reopened = Signal()

    def __init__(
        self,
        x_name,
        y_names=None,
        *,
        legend_keys=None,
        xlim=None,
        ylim=None,
        epoch="run",
        parent=None,
        namespace=None,
        ylabel="",
        xlabel="",
        ylabel2="",
        title="",
        stream_name="primary",
        fits=None,
        do_plot=True,
        multi_stream=False,
        y_axes=None,
        logX=False,
        logY=False,
        logY2=False,
        maxlen=np.inf,
        use_bluesky=True,
        labels=(),
        first_hidden=None,
        **kwargs,
    ):
        super().__init__(parent=parent)
        self.plot_widget = pg.PlotWidget()
        self.x_name = x_name
        self.y_names = y_names or y_axes.keys()
        self.stream_name = stream_name
        self.fits = fits or []
        self.eva = Evaluator(namespace=namespace)
        self.liveFits = []
        self.liveFitPlots = []
        self.ax2_viewbox = None
        self.setWindowTitle(title or f"{x_name} vs. {y_names[0]}")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))

        ax2 = None
        plotItem = self.plot_widget.getPlotItem()
        if y_axes and 2 in y_axes.values():
            self.ax2_viewbox = pg.ViewBox()
            plotItem.scene().addItem(self.ax2_viewbox)
            ax2 = plotItem.getAxis("right")
            self.ax2_viewbox.setXLink(plotItem)
            ax2 = pg.AxisItem("right")
            ax2.setLabel(ylabel2 or self.y_names[0])
            ax2.linkToView(self.ax2_viewbox)
            plotItem.layout.addItem(ax2, 2, 3)

            # This makes the second y-axis move with the viewbox of the main plot
            def updateViews():
                self.ax2_viewbox.setGeometry(plotItem.vb.sceneBoundingRect())
                self.ax2_viewbox.linkedViewChanged(plotItem.vb, self.ax2_viewbox.XAxis)

            updateViews()
            plotItem.vb.sigResized.connect(updateViews)

        # create the fits
        for fit in self.fits:
            if fit["use_custom_func"]:
                model = lmfit.models.ExpressionModel(fit["custom_func"])
                label = "custom"
            else:
                model = lmfit.models.lmfit_models[fit["predef_func"]]()
                label = fit["predef_func"]
            if fit["guess_params"]:
                init_guess = None
            else:
                init_guess = {}
                for i, param in enumerate(fit["initial_params"]["name"]):
                    init_guess[param] = fit["initial_params"]["initial value"][i]
            upper = {}
            lower = {}
            for i, param in enumerate(fit["initial_params"]["name"]):
                upper[param] = fit["initial_params"]["upper bound"][i] or np.inf
                lower[param] = fit["initial_params"]["lower bound"][i] or -np.inf
            params = model.make_params()
            if init_guess:
                for i, param in enumerate(fit["initial_params"]["name"]):
                    params[param].set(
                        init_guess[param], min=lower[param], max=upper[param]
                    )
            else:
                for i, param in enumerate(fit["initial_params"]["name"]):
                    params[param].set(min=lower[param], max=upper[param])
            name = f'{label}_{fit["y"]}_v_{fit["x"]}'
            name = replace_name(name)
            add_data = fit["additional_data"] or {}
            livefit = LiveFit_Eva(
                model,
                fit["y"],
                {"x": fit["x"]},
                self.eva,
                init_guess,
                name=name,
                additional_data=add_data,
                params=params,
                stream_name=stream_name,
            )
            self.liveFits.append(livefit)
            if y_axes and y_axes[fit["y"]] == 2:
                viewbox = self.ax2_viewbox
            else:
                viewbox = self.plot_widget.getPlotItem().vb
            self.liveFitPlots.append(
                LiveFitPlot(
                    livefit, viewbox, plotItem, display_values=fit["display_values"]
                )
            )

        # create the main plot, either with or without using bluesky callbacks
        if use_bluesky:
            self.livePlot = LivePlot(
                x_name=x_name,
                y_names=y_names,
                maxlen=maxlen,
                multi_stream=multi_stream,
                stream_name=stream_name,
                evaluator=self.eva,
                y_axes=y_axes,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                epoch=epoch,
                plot_item=self.plot_widget.getPlotItem(),
                ax2_viewbox=self.ax2_viewbox,
                ax2_axis=ax2,
                fitPlots=self.liveFitPlots,
            )
        else:
            self.livePlot = LivePlot_NoBluesky(
                self.plot_widget.getPlotItem(),
                ax2_viewbox=self.ax2_viewbox,
                ax2_axis=ax2,
                xlabel=xlabel,
                ylabel=ylabel,
                ylabel2=ylabel2,
                y_axes=y_axes,
                labels=labels,
                first_hidden=first_hidden,
            )
        self.livePlot.plotItem.showGrid(True, True)
        self.livePlot.new_data_signal.connect(self.show_again)
        self.livePlot.setup_done_signal.connect(self.make_toolbar)
        self.toolbar = None
        self.pushButton_show_options = QPushButton("Show Options")
        self.pushButton_show_options.clicked.connect(self.show_options)
        self.pushButton_clear = QPushButton("Clear Plot")
        self.pushButton_clear.clicked.connect(self.clear_plot)
        self.plot_options = Plot_Options(self, self.livePlot)
        self.plot_options.checkBox_log_x.setChecked(logX)
        self.plot_options.checkBox_log_y.setChecked(logY)
        self.plot_options.checkBox_log_y2.setChecked(logY2)
        self.options_open = False
        label_n_data = QLabel("# data points:")
        self.lineEdit_n_data = QLineEdit(str(maxlen))
        self.lineEdit_n_data.returnPressed.connect(self.change_maxlen)
        label_skip_n_points = QLabel("skip n points:")
        self.lineEdit_skip_n_points = QLineEdit("0")
        self.lineEdit_skip_n_points.returnPressed.connect(self.change_skip_n_points)
        skip_tool_tip = "Skip n points in the plot before plotting the next value. This may be useful for long measurements to speed up plotting and free up memory."
        label_skip_n_points.setToolTip(skip_tool_tip)
        self.lineEdit_skip_n_points.setToolTip(skip_tool_tip)
        self.setLayout(QGridLayout())
        self.layout().addWidget(self.plot_widget, 0, 1, 1, 6)
        self.layout().addWidget(self.pushButton_show_options, 2, 1)
        self.layout().addWidget(self.pushButton_clear, 2, 2)
        self.layout().addWidget(label_n_data, 2, 5)
        self.layout().addWidget(self.lineEdit_n_data, 2, 6)
        self.layout().addWidget(label_skip_n_points, 2, 3)
        self.layout().addWidget(self.lineEdit_skip_n_points, 2, 4)
        self.layout().addWidget(self.plot_options, 0, 0, 3, 1)
        self.plot_options.hide()
        self.setMinimumSize(500, 400)
        self.adjustSize()
        place_widget(self)
        self.change_maxlen()

    def show_again(self):
        if not self.isVisible():
            self.show()
            self.reopened.emit()

    def make_toolbar(self):
        """Creates the toolbar for the plot widget. This toolbar is based on the context menu of the plot. The View All is connected with `auto_range`."""
        self.toolbar = QMenuBar()
        viewbox = self.plot_widget.getPlotItem().getViewBox()
        menu = viewbox.menu
        scene = viewbox.scene()
        # dummy mouseclickevent
        press_event = QGraphicsSceneMouseEvent(QEvent.GraphicsSceneMousePress)
        event = MouseClickEvent(press_event)
        scene.addParentContextMenus(viewbox, menu, event)
        actions = menu.actions()
        for action in actions:
            if action.text() == "View All":
                action.triggered.connect(self.auto_range)
            self.toolbar.addAction(action)
        self.layout().addWidget(self.toolbar, 1, 1, 1, 6)
        self.plot_options.set_log()

    def auto_range(self):
        """Also call the auto range for the second y-axis if it exists."""
        self.livePlot.plotItem.vb.autoRange()
        if self.ax2_viewbox:
            self.ax2_viewbox.autoRange()

    def change_maxlen(self):
        """
        Changes the maximum number of data points to show in the plot. Reads the value from the line edit and sets it as the new maximum length.
        """
        text = self.lineEdit_n_data.text()
        if not text or text.lower() in ["none", "inf", "np.inf"]:
            maxlen = np.inf
        else:
            try:
                maxlen = int(text)
            except ValueError:
                return
        self.livePlot.change_maxlen(maxlen)

    def change_skip_n_points(self):
        text = self.lineEdit_skip_n_points.text()
        if not text:
            skip_n_points = 0
        else:
            try:
                skip_n_points = int(text)
            except ValueError:
                return
        self.livePlot.change_skip_n_points(skip_n_points)

    def show_options(self):
        """
        Shows or hides the plot options. If the options are shown, the button text is changed to "Hide Options", otherwise it is changed to "Show Options".
        """
        if self.options_open:
            self.pushButton_show_options.setText("Show Options")
            self.options_open = False
            self.plot_options.hide()
            self.adjustSize()
        else:
            self.pushButton_show_options.setText("Hide Options")
            self.options_open = True
            self.plot_options.show()
            self.plot_options.update_plot_items()

    def clear_plot(self):
        """
        Clears the plot and the fits.
        """
        self.livePlot.clear_plot()
        for fit in self.liveFitPlots:
            fit.clear_plot()

    def closeEvent(self, event):
        """
        Emits the closing signal when the widget is closed.
        """
        self.closing.emit()
        super().closeEvent(event)


class Plot_Options(Ui_Plot_Options, QWidget):
    """
    Class for the plot options widget. This widget is used to change the appearance of the plot, such as the color, linestyle, marker, and log-scale of the axes.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget, by default None
    livePlot : LivePlot
        The LivePlot object connected to the plot widget, by default None
    """

    def __init__(self, parent=None, livePlot=None):
        super().__init__(parent=parent)
        self.setupUi(self)
        self.livePlot = livePlot
        self.all_items = {}
        self.items_in_2 = []
        self.checkBox_log_x.setChecked(self.livePlot.plotItem.getAxis("bottom").logMode)
        self.checkBox_log_y.setChecked(self.livePlot.plotItem.getAxis("left").logMode)
        if self.livePlot.ax2_viewbox:
            self.checkBox_log_y2.setChecked(
                self.livePlot.plotItem.getAxis("right").logMode
            )
        self.checkBox_log_x.stateChanged.connect(self.set_log)
        self.checkBox_log_y.stateChanged.connect(self.set_log)
        self.checkBox_log_y2.stateChanged.connect(self.set_log)
        self.checkBox_use_abs_x.stateChanged.connect(self.set_log)
        self.checkBox_use_abs_y.stateChanged.connect(self.set_log)
        self.checkBox_use_abs_y2.stateChanged.connect(self.set_log)

    def update_plot_items(self):
        """
        Updates the items in the table widget to match the items in the plot.
        """
        self.all_items.clear()
        self.items_in_2.clear()
        items_viewbox_1 = self.livePlot.plotItem.vb.allChildItems()
        # get all PlotDataItems in the first y-axis
        for item in items_viewbox_1:
            if isinstance(item, pg.PlotDataItem):
                self.all_items[item.name()] = item
        # get all PlotDataItems in the second y-axis
        if self.livePlot.ax2_viewbox:
            items_viewbox_2 = self.livePlot.ax2_viewbox.allChildItems()
            for item in items_viewbox_2:
                if isinstance(item, pg.PlotDataItem):
                    self.all_items[item.name()] = item
                    self.items_in_2.append(item.name())
        self.tableWidget.clear()
        self.tableWidget.setColumnCount(4)
        self.tableWidget.setMinimumWidth(400)
        self.tableWidget.setHorizontalHeaderLabels(
            ["Name", "marker", "linestyle", "color"]
        )
        self.tableWidget.verticalHeader().setHidden(True)
        self.tableWidget.setRowCount(len(self.all_items))
        for i, (name, item) in enumerate(
            sorted(self.all_items.items(), key=lambda x: x[0].lower())
        ):
            self.tableWidget.setItem(i, 0, QTableWidgetItem(name))

            markerwidge = QComboBox()
            markerwidge.addItems(symbols.keys())
            marker = item.opts["symbol"]
            if marker in symbols.values():
                marker = list(symbols.keys())[list(symbols.values()).index(marker)]
            markerwidge.setCurrentText(marker)
            markerwidge.currentTextChanged.connect(
                lambda text, n=i: self.change_symbol(text, n)
            )
            self.tableWidget.setCellWidget(i, 1, markerwidge)

            linestylewidge = QComboBox()
            linestylewidge.addItems(linestyles.keys())
            linestyle = item.opts["pen"].style()
            if linestyle in linestyles.values():
                linestyle = list(linestyles.keys())[
                    list(linestyles.values()).index(linestyle)
                ]
            linestylewidge.setCurrentText(linestyle)
            linestylewidge.currentTextChanged.connect(
                lambda text, item=item: item.setPen(
                    pg.mkPen(
                        color=item.opts["pen"].color(),
                        width=item.opts["pen"].width(),
                        style=linestyles[text],
                    )
                )
            )
            self.tableWidget.setCellWidget(i, 2, linestylewidge)

            color = item.opts["pen"].color()
            colorwidge = QPushButton(color.name())
            colorwidge.clicked.connect(lambda n=i: self.change_color(n))
            self.tableWidget.setCellWidget(i, 3, colorwidge)

    def change_symbol(self, symbol, row):
        """
        Changes the symbol of a PlotDataItem connected to the specified `row`.

        Parameters
        ----------
        symbol : str
            The new symbol
        row : int
            The row of the item in the table widget
        """
        name = self.tableWidget.item(row, 0).text()
        self.change_color(row, just_update=True)
        self.all_items[name].setSymbol(symbols[symbol])

    def change_color(self, row, just_update=False):
        """
        Changes the color of a PlotDataItem connected to the specified `row`.

        Parameters
        ----------
        row : int
            The row of the item in the table widget
        just_update : bool, (default: False)
            If True, only updates the color of the item to the one already set. If False, opens a color dialog to choose a new color.
        """
        item = self.tableWidget.cellWidget(row, 3)
        if just_update:
            color = QColor(item.text())
        else:
            color = QColorDialog.getColor()
            if not color.isValid():
                return
        item.setText(color.name())
        name = self.tableWidget.item(row, 0).text()
        self.all_items[name].opts["pen"].setColor(color)
        self.all_items[name].setSymbolPen(pg.mkPen(color=color))
        self.all_items[name].setSymbolBrush(pg.mkBrush(color=color))

    def set_log(self):
        """
        Sets the log-scale of the axes according to the checkboxes.
        """
        x = self.checkBox_log_x.isChecked()
        self.checkBox_use_abs_x.setEnabled(x)
        y = self.checkBox_log_y.isChecked()
        self.checkBox_use_abs_y.setEnabled(y)
        y2 = self.checkBox_log_y2.isChecked()
        self.checkBox_use_abs_y2.setEnabled(y2)
        self.livePlot.use_abs["x"] = self.checkBox_use_abs_x.isChecked()
        self.livePlot.use_abs["y"] = self.checkBox_use_abs_y.isChecked()
        self.livePlot.use_abs["y2"] = self.checkBox_use_abs_y2.isChecked()
        for name, item in self.all_items.items():
            if name in self.items_in_2:
                item.setLogMode(x, y2)
            else:
                item.setLogMode(x, y)
        if self.livePlot.ax2_viewbox:
            self.livePlot.plotItem.getAxis("right").setLogMode(x, y2)
            self.livePlot.ax2_viewbox.setLogMode(x, y2)
            self.livePlot.ax2_axis.setLogMode(y2)
            self.livePlot.ax2_viewbox.enableAutoRange()
        self.livePlot.plotItem.setLogMode(x, y)
        self.livePlot.update_plot()


class LivePlot(QObject, CallbackBase):
    """
    Bluesky callback class for live plotting. This class is used to update the plot with new data.

    Parameters
    ----------
    x_name : str
        The name of the x-axis variable
    y_names : List[str]
        The name(s) of the y-axis variable(s)
    plot_item : pg.PlotItem
        The plot item to plot the data on
    maxlen : [int, np.inf], (default: np.inf)
        The maximum number of data points to show
    multi_stream : bool, (default: False)
        Whether to use multiple streams. If True, all streams including the `stream_name` are used.
    evaluator : Evaluator
        The evaluator object used to evaluate expressions.
    stream_name : str, (default: 'primary')
        The name of the bluesky stream to use for the plot.
    y_axes : Dict[str, int], (default: None)
        The y-axis to use for each y_name, the ints should be 1 or 2, if 2, the respective y-value is plotted on the right axis
    title : str, (default: '')
        The title of the plot
    xlabel : str, (default: '')
        The x-axis label, if empty the x_name is used
    ylabel : str, (default: '')
        The y-axis label, if empty the first y_name is used
    epoch : {'run', 'unix'}, (default: 'run')
        If 'run' t=0 is the time recorded in the RunStart document. If 'unix', t=0 is 1 Jan 1970 ("the UNIX epoch").
    ax2_viewbox : pg.ViewBox, (default: None)
        The viewbox for the second y-axis
    ax2_axis : pg.AxisItem, (default: None)
        The axis for the second y-axis
    fitPlots : List[LiveFitPlot], (default: [])
        The fit plots connected to the plot.
    """

    new_data_signal = Signal()
    setup_done_signal = Signal()

    def __init__(
        self,
        x_name,
        y_names,
        plot_item,
        *,
        maxlen=np.inf,
        multi_stream=False,
        evaluator=None,
        stream_name="primary",
        y_axes=None,
        title="",
        xlabel=None,
        ylabel=None,
        epoch="run",
        ax2_viewbox=None,
        ax2_axis=None,
        fitPlots=None,
        **kwargs,
    ):
        CallbackBase.__init__(self)
        QObject.__init__(self)
        self.__teleporter = Teleporter()
        self.__teleporter.name_doc_escape.connect(handle_teleport)
        self.plotItem = plot_item
        self.__setup_lock = threading.Lock()
        self.__setup_event = threading.Event()
        self.use_abs = {"x": False, "y": False, "y2": False}
        self.setup_is_done = False

        def setup():
            # this is the setup function, it is called when the first event is received
            nonlocal y_names, x_name, title, xlabel, ylabel, epoch
            with self.__setup_lock:
                if self.__setup_event.is_set():
                    return
                self.__setup_event.set()
            # set the labels
            self.plots = {}
            if x_name is not None:
                self.x, *others = get_obj_fields([x_name])
            else:
                self.x = "seq_num"
            self.ys = get_obj_fields(y_names)
            self.plotItem.setLabel("bottom", xlabel or x_name or "sequence #")
            self.plotItem.setLabel("left", ylabel or self.ys[0])
            if title:
                self.plotItem.setTitle(title)
            self._epoch_offset = None
            self._epoch = epoch
            self.setup_is_done = True

        self.x_data = ListDeque_skip(maxlen=maxlen)
        self.y_data = {}
        self.y_axes = y_axes or {}
        self.maxlen = maxlen
        self.stream_name = stream_name
        self.eva = evaluator
        self.y_names = y_names
        self.ys = get_obj_fields(y_names)
        for y in y_names:
            self.y_data[y] = ListDeque_skip(maxlen=maxlen)
        self.current_plots = {}
        self.legend = None
        self.desc = []
        self.multi_stream = multi_stream
        self.ax2_viewbox = ax2_viewbox
        self.ax2_axis = ax2_axis
        self.__setup = setup
        self.fitPlots = fitPlots or []
        for fit in self.fitPlots:
            fit.parent_plot = self
        self.descs_fit_readying = {}
        self.line_number = 0
        self.n_plots = 0

    def __call__(self, name, doc, *, escape=False):
        """
        The call method of the callback. This method is called when a new event is received. If `__teleporter` is set, the event is sent to the `__teleporter`, otherwise the event is processed directly. The teleporter is necessary to send the event to a different thread, since Qt objects can only be accessed from the thread they were created in.

        Parameters
        ----------
        name : str
            The name of the event
        doc : dict
            The event document
        escape : bool, (default: False)
            If True, the event is always processed directly, otherwise it is sent to the teleporter
        """
        if not escape and self.__teleporter is not None:
            self.__teleporter.name_doc_escape.emit(name, doc, self)
        else:
            return CallbackBase.__call__(self, name, doc)

    def add_plot(self, y):
        i = self.ys.index(y)
        color = colors[self.n_plots % len(colors)]
        if y in self.y_axes and self.y_axes[y] == 2:
            self.current_plots[y] = plot = pg.PlotDataItem(
                [],
                [],
                label=self.y_names[i],
                name=self.y_names[i],
                symbol="o",
                symbolPen=pg.mkPen(color=color),
                symbolBrush=pg.mkBrush(color=color),
                pen=pg.mkPen(color=color, width=2, style=linestyles["none"]),
            )
            self.ax2_viewbox.addItem(plot)
        else:
            self.current_plots[y] = self.plotItem.plot(
                [],
                [],
                label=self.y_names[i],
                name=self.y_names[i],
                symbol="o",
                symbolPen=pg.mkPen(color=color),
                symbolBrush=pg.mkBrush(color=color),
                pen=pg.mkPen(color=color, width=2, style=linestyles["none"]),
            )
        self.n_plots += 1

    def start(self, doc):
        """
        This method is called when the RunStart document is received. It sets up the plot and the fits.

        Parameters
        ----------
        doc : dict
            The RunStart document
        """
        self.__setup()
        for i, y in enumerate(self.ys):
            try:
                self.add_plot(y)
            except Exception as e:
                print(e)
        self.legend = pg.LegendItem(
            offset=(1, 1), horSpacing=20, verSpacing=-5, pen="w" if dark_mode else "k"
        )
        self.legend.setParentItem(self.plotItem.vb)
        for plot in self.current_plots.values():
            self.legend.addItem(plot, plot.name())
        self.setup_done_signal.emit()
        for fit in self.fitPlots:
            fit.start(doc)
        self.eva.start(doc)

    def descriptor(self, doc):
        """
        This method is called when a new descriptor document is received. If the descriptor is relevant, (compared with `self.stream_name`), the uid of the descriptor is added to the list of relevant descriptors. If the descriptor is a fit descriptor, the fit stream is added to the list of fits.

        Parameters
        ----------
        doc : dict
            The descriptor document
        """
        if not self.setup_is_done:
            self.start(doc)
        if doc["name"] == self.stream_name:
            self.desc.append(doc["uid"])
        elif doc["name"].startswith(f"{self.stream_name}_fits_readying_"):
            for fit in self.fitPlots:
                if (
                    doc["name"]
                    == f"{self.stream_name}_fits_readying_{fit.livefit.name}"
                ):
                    self.descs_fit_readying[doc["uid"]] = fit
        elif self.multi_stream and doc["name"].startswith(self.stream_name):
            self.desc.append(doc["uid"])

    def event(self, doc):
        """
        This method is called when a new event document is received. If the descriptor of the event is not in the list of relevant descriptors, the event is ignored. Otherwise, the data is extracted from the event and added to the plot.

        Parameters
        ----------
        doc : dict
            The event document
        """
        if isinstance(doc, QEvent):
            try:
                pg.PlotWidget.event(self, doc)
            except Exception as e:
                print(e)
            return
        if doc["descriptor"] not in self.desc:
            if doc["descriptor"] in self.descs_fit_readying:
                self.descs_fit_readying[doc["descriptor"]].get_ready()
            return
        try:
            new_x = doc["data"][self.x]
        except KeyError:
            if self.x in ("time", "seq_num"):
                new_x = doc[self.x]
            else:
                if not self.eva.is_to_date(doc["time"]):
                    self.eva.event(doc)
                new_x = self.eva.eval(self.x, do_not_reraise=True)
        if self.x == "time" and self._epoch == "run":
            new_x -= self._epoch_offset

        new_y = {}
        for y in self.ys:
            try:
                new_y[y] = doc["data"][y]
            except KeyError:
                if not self.eva.is_to_date(doc["time"]):
                    self.eva.event(doc)
                new_y[y] = self.eva.eval(y, do_not_reraise=True)
        self.update_caches(new_x, new_y)
        for fit in self.fitPlots:
            fit.event(doc)
        self.update_plot()

    def update_caches(self, new_x, new_y):
        for y in new_y:
            self.y_data[y].append(new_y[y])
        self.x_data.append(new_x)

    def update_plot(self):
        if self.plotItem.getAxis("bottom").logMode and self.use_abs["x"]:
            plot_x = np.abs(self.x_data)
        else:
            plot_x = self.x_data
        plot_x = np.asarray(plot_x)
        if plot_x.ndim > 1:
            plot_x = plot_x[-1]
        for y in self.ys:
            y_abs = False
            y2_abs = False
            if self.plotItem.getAxis("left").logMode and self.use_abs["y"]:
                y_abs = True
            if (
                self.ax2_viewbox
                and self.plotItem.getAxis("right").logMode
                and self.use_abs["y2"]
            ):
                y2_abs = True
            if self.y_axes.get(y, 1) == 2:
                plot_y = np.abs(self.y_data[y]) if y2_abs else self.y_data[y]
            else:
                plot_y = np.abs(self.y_data[y]) if y_abs else self.y_data[y]
            plot_y = np.asarray(plot_y)
            if plot_y.ndim > 1:
                plot_y = plot_y[-1]
            if not y in self.current_plots:
                self.add_plot(y)
            self.current_plots[y].setData(plot_x, plot_y)
        self.new_data_signal.emit()

    def stop(self, doc):
        if not self.x_data:
            print(
                f"LivePlot did not get any data that corresponds to the x axis. {self.x}"
            )
        for y in self.y_data:
            if not self.y_data[y]:
                print(f"LivePlot did not get any data for {y}")
            if len(self.y_data[y]) != len(self.x_data):
                print(f"LivePlot has a length mismatch for {y}")
        for fit in self.fitPlots:
            fit.stop(doc)

    def clear_plot(self):
        for y in self.current_plots:
            self.current_plots[y].setData([], [])
        self.x_data = ListDeque_skip(maxlen=self.maxlen)
        for y in self.y_data:
            self.y_data[y] = ListDeque_skip(maxlen=self.maxlen)
        for fit in self.fitPlots:
            fit.clear_plot()
        self.update_plot()

    def change_maxlen(self, maxlen):
        self.maxlen = maxlen
        self.x_data.change_maxlen(maxlen)
        for y in self.y_data:
            self.y_data[y].change_maxlen(maxlen)

    def change_skip_n_points(self, n):
        self.x_data.skip_n_points = n
        for y in self.y_data:
            self.y_data[y].skip_n_points = n


class LiveFitPlot(CallbackBase):
    def __init__(
        self,
        livefit,
        viewbox,
        plotItem,
        *,
        num_points=100,
        display_values=False,
        **kwargs,
    ):
        super().__init__()
        self.__teleporter = Teleporter()
        self.__teleporter.name_doc_escape.connect(handle_teleport)
        if len(livefit.independent_vars) != 1:
            raise NotImplementedError(
                "LiveFitPlot supports models with one independent variable only."
            )

        self.viewbox = viewbox
        self.plotItem = plotItem
        self.livefit = livefit
        self.display_values = display_values
        self.num_points = num_points
        self.kwargs = kwargs

        (self.__x_key,) = livefit.independent_vars.keys()
        self._has_been_run = False
        livefit.parent_plot = self
        self.x_data = []
        self.y_data = []
        self.plot = None
        self.x = None
        self.line_position = None
        self.text_objects = []
        self.color = "w"

        self.__setup_lock = threading.Lock()
        self.__setup_event = threading.Event()

        def setup():
            nonlocal livefit
            with self.__setup_lock:
                if self.__setup_event.is_set():
                    return
                self.__setup_event.set()
            (x,) = livefit.independent_vars.values()
            if x is not None:
                self.x, *others = get_obj_fields([x])
            else:
                self.x = "seq_num"
            y = livefit.y
            self.y, *others = get_obj_fields([y])

        self.__setup = setup

    def start(self, doc):
        self.__setup()
        self.x_data, self.y_data = [], []
        self.color = colors[self.parent_plot.n_plots % len(colors)]
        self.plot = pg.PlotDataItem(
            [],
            [],
            pen=pg.mkPen(color=self.color, width=2, style=linestyles["solid"]),
            name=self.livefit.name,
            symbol=None,
            symbolPen=pg.mkPen(color=self.color),
            symbolBrush=pg.mkBrush(color=self.color),
        )
        self.viewbox.addItem(self.plot)
        self.livefit.start(doc)
        (self.x,) = self.livefit.independent_vars.keys()
        self.parent_plot.n_plots += 1

    def get_ready(self):
        """Passes the command to the `_livefit`"""
        self.livefit.get_ready()

    def event(self, doc):
        """Passes the event to the `livefit`"""
        self.livefit.event(doc)

    def fit_has_result(self):
        if self.livefit.result is not None:
            x_data = self.livefit.independent_vars_data[self.__x_key]
            x_points = np.linspace(np.min(x_data), np.max(x_data), self.num_points)
            kwargs = {self.__x_key: x_points}
            kwargs.update(self.livefit.result.values)
            self.y_data = self.livefit.result.model.eval(**kwargs)
            self.x_data = x_points
            kwargs.update(self.livefit.result.init_values)
            self.update_plot()

    def clear_plot(self):
        if self.plot:
            self.plot.setData([], [])
        for text in self.text_objects:
            self.viewbox.removeItem(text)
        self.text_objects = []

    def update_plot(self):
        self.plot.setData(self.x_data, self.y_data)
        if self.display_values:
            for text in self.text_objects:
                self.viewbox.removeItem(text)
            self.text_objects.clear()
            legend = self.parent_plot.legend
            y0 = 0
            if legend is not None:
                legendRect = legend.boundingRect()
                legendPos = legend.scenePos()
                y0 = legendPos.y() + legendRect.height()
            vals = self.livefit.result.values
            variables = self.livefit.result.var_names
            if self.line_position is None:
                self.line_position = self.parent_plot.line_number
                self.parent_plot.line_number += len(variables)
            for i, (name, value) in enumerate(vals.items()):
                if name not in variables:
                    continue
                if self.livefit.result.covar is not None:
                    error = np.sqrt(self.livefit.result.covar[i, i])
                    text = pg.TextItem(
                        f"{name}: {value:.3e} ± {error:.3e}", color=self.color
                    )
                else:
                    text = pg.TextItem(f"{name}: {value:.3e}", color=self.color)
                text.setParentItem(self.plotItem.vb)
                text.setPos(5, (i + self.line_position) * 20 + y0)
                self.text_objects.append(text)
        self.parent_plot.update_plot()

    def __call__(self, name, doc, *, escape=False):
        if not escape and self.__teleporter is not None:
            self.__teleporter.name_doc_escape.emit(name, doc, self)
        else:
            return CallbackBase.__call__(self, name, doc)


class Teleporter(QObject):
    name_doc_escape = Signal(str, dict, object)


def handle_teleport(name, doc, obj):
    obj(name, doc, escape=True)


class PlotWidget_2D(QWidget):
    """ """

    closing = Signal()
    reopened = Signal()

    def __init__(
        self,
        x_name,
        y_name,
        z_name,
        parent=None,
        namespace=None,
        xlabel="",
        ylabel="",
        zlabel="",
        title="",
        maxlen=np.inf,
        stream_name="primary",
        **kwargs,
    ):
        super().__init__(parent=parent)
        self.setWindowTitle(
            title or f"{zlabel or z_name} vs. {xlabel or x_name}, {ylabel or y_name}"
        )
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))
        self.graphics_layout = pg.GraphicsLayoutWidget()
        self.plot = self.graphics_layout.addPlot(col=0, row=0)
        self.x_name = x_name
        self.y_name = y_name
        self.z_name = z_name
        self.plot.setLabel("bottom", xlabel or x_name)
        self.plot.setLabel("left", ylabel or y_name)
        self.stream_name = stream_name

        self.toolbar = None
        eva = Evaluator(namespace=namespace)
        self.livePlot = LivePlot_2D(
            x_name,
            y_name,
            z_name,
            plotItem=self.plot,
            graphics_layout=self.graphics_layout,
            cmap="viridis",
            evaluator=eva,
            stream_name=stream_name,
            **kwargs,
        )
        self.livePlot.new_data.connect(self.show_again)
        label_n_data = QLabel("# data points:")
        self.lineEdit_n_data = QLineEdit(str(maxlen))
        self.lineEdit_n_data.returnPressed.connect(self.change_maxlen)
        self.pushButton_clear = QPushButton("Clear Plot")
        self.pushButton_clear.clicked.connect(self.clear_plot)

        self.setLayout(QGridLayout())
        self.layout().addWidget(self.graphics_layout, 0, 0, 1, 3)
        self.layout().addWidget(self.pushButton_clear, 2, 0)
        self.layout().addWidget(label_n_data, 2, 1)
        self.layout().addWidget(self.lineEdit_n_data, 2, 2)
        self.make_toolbar()
        self.adjustSize()
        place_widget(self)
        self.change_maxlen()

    def change_maxlen(self):
        """
        Changes the maximum number of data points to show in the plot. Reads the value from the line edit and sets it as the new maximum length.
        """
        text = self.lineEdit_n_data.text()
        if not text or text.lower() in ["none", "inf", "np.inf"]:
            maxlen = np.inf
        else:
            try:
                maxlen = int(text)
            except ValueError:
                return
        self.livePlot.change_maxlen(maxlen)

    def show_again(self):
        if not self.isVisible():
            self.show()
            self.reopened.emit()

    def make_toolbar(self):
        self.toolbar = QMenuBar()
        viewbox = self.plot.getViewBox()
        menu = viewbox.menu
        scene = viewbox.scene()
        # dummy mouseclickevent
        press_event = QGraphicsSceneMouseEvent(QEvent.GraphicsSceneMousePress)
        event = MouseClickEvent(press_event)
        scene.addParentContextMenus(viewbox, menu, event)
        actions = menu.actions()
        for action in actions:
            self.toolbar.addAction(action)
        self.layout().addWidget(self.toolbar, 1, 0, 1, 3)

    def clear_plot(self):
        self.livePlot.clear_plot()

    def closeEvent(self, event):
        self.closing.emit()
        super().closeEvent(event)


class LivePlot_2D(QObject, CallbackBase):
    """ """

    new_data = Signal()

    def __init__(
        self,
        x,
        y,
        z,
        plotItem,
        graphics_layout,
        *,
        cmap="viridis",
        evaluator=None,
        stream_name="primary",
        **kwargs,
    ):
        CallbackBase.__init__(self)
        QObject.__init__(self)
        self.__teleporter = Teleporter()
        self.__teleporter.name_doc_escape.connect(handle_teleport)
        self.x = x
        self.y = y
        self.z = z
        self.x_data, self.y_data, self.z_data = [], [], []
        self.cmap = pg.colormap.get(cmap)
        self.eva = evaluator
        self.stream_name = stream_name
        self.kwargs = kwargs
        self.plotItem = plotItem
        self.graphics_layout = graphics_layout
        self.scatter_plot = None
        self.image = None
        self.hist = None
        self.desc = None
        self._epoch_offset = None
        self._epoch = "run"
        self._minx = self._miny = np.inf
        self._maxx = self._maxy = -np.inf

    def __call__(self, name, doc, *, escape=False):
        if not escape and self.__teleporter is not None:
            self.__teleporter.name_doc_escape.emit(name, doc, self)
        else:
            return CallbackBase.__call__(self, name, doc)

    def start(self, doc):
        self.x_data.clear()
        self.y_data.clear()
        self.z_data.clear()
        self.scatter_plot = pg.ScatterPlotItem(self.x_data, self.y_data)
        self.dummy_image = pg.ImageItem()
        self.color_bar = pg.ColorBarItem(label=self.z, interactive=False)
        self.color_bar.setColorMap(self.cmap)
        self.color_bar.setImageItem(self.dummy_image)
        # self.color_bar.sigLevelsChanged.connect(self.update_scatter)
        self.plotItem.addItem(self.scatter_plot)
        self.image = pg.ImageItem()
        self.hist = pg.HistogramLUTItem()
        self.hist.setImageItem(self.image)
        self.hist.autoHistogramRange()
        self.hist.gradient.setColorMap(self.cmap)
        self.hist.axis.setLabel(self.z)
        self.graphics_layout.addItem(self.hist, row=0, col=1)
        self.graphics_layout.addItem(self.color_bar, row=0, col=2)
        self.plotItem.addItem(self.image)
        self._epoch_offset = None
        self._epoch = "run"

    def update_scatter(self):
        try:
            current_map = self.color_bar.getColorMap()
            colors = current_map.map(self.z_normed)
            self.scatter_plot.setData(x=self.x_data, y=self.y_data, brush=colors)
        except Exception as e:
            pass

    def descriptor(self, doc):
        if doc["name"] == self.stream_name:
            self.desc = doc["uid"]

    def event(self, doc):
        if doc["descriptor"] != self.desc:
            return
        try:
            x = doc["data"][self.x]
        except KeyError:
            if self.x in ("time", "seq_num"):
                x = doc[self.x]
            else:
                if not self.eva.is_to_date(doc["time"]):
                    self.eva.event(doc)
                x = self.eva.eval(self.x, do_not_reraise=True)
        try:
            y = doc["data"][self.y]
        except KeyError:
            if not self.eva.is_to_date(doc["time"]):
                self.eva.event(doc)
            y = self.eva.eval(self.y, do_not_reraise=True)
        try:
            z = doc["data"][self.z]
        except KeyError:
            if not self.eva.is_to_date(doc["time"]):
                self.eva.event(doc)
            z = self.eva.eval(self.z, do_not_reraise=True)
        self.update(x, y, z)
        self.new_data.emit()

    def make_colormesh(self, x_shape=None, y_shape=None):
        if x_shape is None and y_shape is None:
            return None
        elif x_shape is not None and y_shape is None:
            y_shape = int(np.array(self.x_data).size / x_shape)
        elif x_shape is None and y_shape is not None:
            x_shape = int(np.array(self.y_data).size / y_shape)
        try:
            x = np.array(self.x_data).reshape((x_shape, y_shape))
            y = np.array(self.y_data).reshape((x_shape, y_shape))
            c = np.array(self.z_data).reshape((x_shape, y_shape))
            return x, y, c
        except Exception as e:
            return None

    def update(self, x, y, z):
        x_shape = None
        y_shape = None
        if np.isscalar(z):
            z = np.array([z])
        if np.isscalar(x):
            x = np.full(z.shape, x)
        if np.isscalar(y):
            y = np.full(z.shape, y)

        self.x_data.extend(x)
        self.y_data.extend(y)
        self.z_data.extend(z)
        x_shape = len(set(self.x_data))
        y_shape = len(set(self.y_data))
        mesh = self.make_colormesh(x_shape, y_shape)
        if mesh:
            x, y, z = mesh
            if not self.x_data:
                return
            self.image.clear()
            self.image.setImage(z)
            self.image.setRect(pg.QtCore.QRectF(x.min(), y.min(), np.ptp(x), np.ptp(y)))
            self.image.setLookupTable(self.cmap.getLookupTable())
            self.scatter_plot.hide()
            self.color_bar.hide()
            self.image.show()
            self.hist.show()
        else:
            self.z_normed = (self.z_data - np.min(self.z_data)) / (
                np.max(self.z_data) - np.min(self.z_data)
            )
            self.dummy_image.setImage(np.array([self.z_data]))
            self.color_bar.setLevels((np.min(self.z_data), np.max(self.z_data)))
            colors = self.cmap.map(self.z_normed)
            self.scatter_plot.setData(x=self.x_data, y=self.y_data, brush=colors)
            self.image.hide()
            self.hist.hide()
            self.scatter_plot.show()
            self.color_bar.show()

    def clear_plot(self):
        self.x_data.clear()
        self.y_data.clear()
        self.z_data.clear()
        self.update(self.x_data, self.y_data, self.z_data)

    def change_maxlen(self, maxlen):
        self.maxlen = maxlen
        if maxlen < np.inf:
            self.x_data = deque(self.x_data, maxlen=maxlen)
            self.y_data = deque(self.y_data, maxlen=maxlen)
            self.z_data = deque(self.z_data, maxlen=maxlen)
        else:
            self.x_data = list(self.x_data)
            self.y_data = list(self.y_data)
            self.z_data = list(self.z_data)


class LivePlot_NoBluesky(QObject):
    new_data_signal = Signal()
    setup_done_signal = Signal()

    def __init__(
        self,
        plotItem,
        xlabel,
        ylabel,
        ylabel2,
        y_axes,
        ax2_viewbox=None,
        ax2_axis=None,
        labels=(),
        first_hidden=None,
    ):
        super().__init__()
        self.plotItem = plotItem
        self.plotItem.setLabel("bottom", xlabel)
        self.plotItem.setLabel("left", ylabel)
        self.ax2_viewbox = ax2_viewbox
        self.ax2_axis = ax2_axis
        if self.ax2_axis:
            self.ax2_axis.setLabel("right", ylabel2)
        self.labels = labels
        self.maxlen = np.inf
        self.x_data = []
        self.y_data = {}
        self.y_axes = y_axes or {}
        self.first_hidden = first_hidden or []
        self.use_abs = {"x": False, "y": False, "y2": False}
        self.current_plots = {}
        self.n_plots = 0
        self.show_plot = True

    def add_data(self, x, ys, add=True):
        if not self.current_plots:
            if len(self.labels) != len(ys):
                self.labels = list(ys.keys())
            for i, y in enumerate(ys):
                try:
                    color = colors[self.n_plots % len(colors)]
                    if y in self.y_axes and self.y_axes[y] == 2:
                        self.current_plots[y] = plot = pg.PlotDataItem(
                            [],
                            [],
                            label=self.labels[i],
                            name=self.labels[i],
                            symbol=None,
                            symbolPen=pg.mkPen(color=color),
                            symbolBrush=pg.mkBrush(color=color),
                            pen=pg.mkPen(
                                color=color,
                                width=2,
                                style=linestyles[
                                    "none" if y in self.first_hidden else "solid"
                                ],
                            ),
                        )
                        self.ax2_viewbox.addItem(plot)
                    else:
                        self.current_plots[y] = self.plotItem.plot(
                            [],
                            [],
                            label=self.labels[i],
                            name=self.labels[i],
                            symbol=None,
                            symbolPen=pg.mkPen(color=color),
                            symbolBrush=pg.mkBrush(color=color),
                            pen=pg.mkPen(
                                color=color,
                                width=2,
                                style=linestyles[
                                    "none" if y in self.first_hidden else "solid"
                                ],
                            ),
                        )
                    if self.maxlen < np.inf:
                        self.y_data[y] = deque(maxlen=self.maxlen)
                    else:
                        self.y_data[y] = []
                    self.n_plots += 1
                except Exception as e:
                    print(e)
            self.setup_done_signal.emit()
        if add:
            self.x_data.append(x)
            for y in ys:
                if y in self.y_data:
                    self.y_data[y].append(ys[y])
        else:
            self.x_data = [x]
            self.y_data.clear()
            for y in ys:
                self.y_data[y] = [ys[y]]
        if self.show_plot:
            self.update_plot()

    def change_maxlen(self, maxlen):
        self.maxlen = maxlen
        if maxlen < np.inf:
            self.x_data = deque(self.x_data, maxlen=maxlen)
            for y in self.y_data:
                self.y_data[y] = deque(self.y_data[y], maxlen=maxlen)
        else:
            self.x_data = list(self.x_data)
            for y in self.y_data:
                self.y_data[y] = list(self.y_data[y])

    def update_plot(self):
        if self.plotItem.getAxis("bottom").logMode and self.use_abs["x"]:
            plot_x = np.abs(self.x_data)
        else:
            plot_x = self.x_data
        plot_x = np.asarray(plot_x)
        if plot_x.ndim > 1:
            plot_x = plot_x[-1]
        for y in self.labels:
            y_abs = False
            y2_abs = False
            if self.plotItem.getAxis("left").logMode and self.use_abs["y"]:
                y_abs = True
            if (
                self.ax2_viewbox
                and self.plotItem.getAxis("right").logMode
                and self.use_abs["y2"]
            ):
                y2_abs = True
            if self.y_axes.get(y, 1) == 2:
                plot_y = np.abs(self.y_data[y]) if y2_abs else self.y_data[y]
            else:
                plot_y = np.abs(self.y_data[y]) if y_abs else self.y_data[y]
            plot_y = np.asarray(plot_y)
            if plot_y.ndim > 1:
                plot_y = plot_y[-1]
            self.current_plots[y].setData(plot_x, plot_y)
        self.new_data_signal.emit()

    def clear_plot(self):
        for y in self.y_data:
            self.current_plots[y].setData([], [])
        self.x_data = []
        for y in self.y_data:
            self.y_data[y] = []
        self.update_plot()



EXTENSION_CONFIG = {
    "required_contexts": [],
    "name": "",
    "version": "0.0",
    "dependencies": {"nomad-camels": ">=0.2.4"},
}


class Extension:
    def stop_extension(self):
        pass



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'general_protocol_settings.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QGridLayout, QHeaderView,
    QLabel, QLineEdit, QPushButton, QRadioButton,
    QSizePolicy, QWidget)

from nomad_camels.ui_widgets.variable_table import VariableTable

class Ui_Protocol_Settings(object):
    def setupUi(self, Protocol_Settings):
        if not Protocol_Settings.objectName():
            Protocol_Settings.setObjectName(u"Protocol_Settings")
        Protocol_Settings.resize(374, 463)
        self.gridLayout = QGridLayout(Protocol_Settings)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.pushButton_add_variable = QPushButton(Protocol_Settings)
        self.pushButton_add_variable.setObjectName(u"pushButton_add_variable")
        self.pushButton_add_variable.setMaximumSize(QSize(27, 27))
        self.pushButton_add_variable.setStyleSheet(u"QPushButton {\n"
"                                background-color:  #4CAF50; \n"
"                                color: white; \n"
"                                border: none; \n"
"                                padding: 0px; \n"
"                                padding-bottom: 5px;\n"
"                                text-align: center; \n"
"                                text-decoration: none; \n"
"                                font-size: 18px; \n"
"                                margin: 2px 2px; \n"
"                                border-radius: 6px;\n"
"								font-weight: bold;\n"
"                            }\n"
"\n"
"            QPushButton:hover {\n"
"                background-color: #45a049;\n"
"            }")

        self.gridLayout.addWidget(self.pushButton_add_variable, 8, 1, 1, 1)

        self.lineEdit_filename = QLineEdit(Protocol_Settings)
        self.lineEdit_filename.setObjectName(u"lineEdit_filename")

        self.gridLayout.addWidget(self.lineEdit_filename, 2, 1, 1, 5)

        self.label = QLabel(Protocol_Settings)
        self.label.setObjectName(u"label")
        self.label.setMaximumSize(QSize(60, 16777215))
        font = QFont()
        font.setBold(True)
        self.label.setFont(font)

        self.gridLayout.addWidget(self.label, 2, 0, 1, 1)

        self.lineEdit_protocol_name = QLineEdit(Protocol_Settings)
        self.lineEdit_protocol_name.setObjectName(u"lineEdit_protocol_name")

        self.gridLayout.addWidget(self.lineEdit_protocol_name, 1, 1, 1, 5)

        self.pushButton_remove_variable = QPushButton(Protocol_Settings)
        self.pushButton_remove_variable.setObjectName(u"pushButton_remove_variable")
        self.pushButton_remove_variable.setMaximumSize(QSize(27, 27))
        self.pushButton_remove_variable.setStyleSheet(u"QPushButton {\n"
"                                background-color:  #FF3333; \n"
"                                color: white; \n"
"                                border: none; \n"
"                                padding: 0px; \n"
"                                padding-bottom: 5px;\n"
"                                text-align: center; \n"
"                                text-decoration: none; \n"
"                                font-size: 18px; \n"
"                                margin: 2px 2px; \n"
"                                border-radius: 6px;\n"
"								font-weight: bold;\n"
"                            }\n"
"\n"
"            QPushButton:hover {\n"
"                background-color: #B22222;\n"
"            }\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout.addWidget(self.pushButton_remove_variable, 8, 2, 1, 1)

        self.widget_2 = QWidget(Protocol_Settings)
        self.widget_2.setObjectName(u"widget_2")
        self.gridLayout_2 = QGridLayout(self.widget_2)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.radioButton_h5_after = QRadioButton(self.widget_2)
        self.radioButton_h5_after.setObjectName(u"radioButton_h5_after")

        self.gridLayout_2.addWidget(self.radioButton_h5_after, 0, 1, 1, 1)

        self.checkBox_json_exp = QCheckBox(self.widget_2)
        self.checkBox_json_exp.setObjectName(u"checkBox_json_exp")

        self.gridLayout_2.addWidget(self.checkBox_json_exp, 1, 1, 1, 1)

        self.checkBox_no_config = QCheckBox(self.widget_2)
        self.checkBox_no_config.setObjectName(u"checkBox_no_config")

        self.gridLayout_2.addWidget(self.checkBox_no_config, 4, 0, 1, 2)

        self.checkBox_csv_exp = QCheckBox(self.widget_2)
        self.checkBox_csv_exp.setObjectName(u"checkBox_csv_exp")

        self.gridLayout_2.addWidget(self.checkBox_csv_exp, 1, 0, 1, 1)

        self.checkBox_NeXus = QCheckBox(self.widget_2)
        self.checkBox_NeXus.setObjectName(u"checkBox_NeXus")
        self.checkBox_NeXus.setChecked(True)

        self.gridLayout_2.addWidget(self.checkBox_NeXus, 3, 0, 1, 2)

        self.radioButton_h5_during = QRadioButton(self.widget_2)
        self.radioButton_h5_during.setObjectName(u"radioButton_h5_during")
        self.radioButton_h5_during.setChecked(True)

        self.gridLayout_2.addWidget(self.radioButton_h5_during, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.widget_2, 4, 0, 1, 6)

        self.variable_table = VariableTable(Protocol_Settings)
        self.variable_table.setObjectName(u"variable_table")
        self.variable_table.verticalHeader().setVisible(False)

        self.gridLayout.addWidget(self.variable_table, 9, 0, 1, 6)

        self.label_title = QLabel(Protocol_Settings)
        self.label_title.setObjectName(u"label_title")
        font1 = QFont()
        font1.setPointSize(12)
        font1.setBold(True)
        self.label_title.setFont(font1)

        self.gridLayout.addWidget(self.label_title, 0, 0, 1, 4)

        self.widget_3 = QWidget(Protocol_Settings)
        self.widget_3.setObjectName(u"widget_3")

        self.gridLayout.addWidget(self.widget_3, 5, 0, 1, 1)

        self.label_4 = QLabel(Protocol_Settings)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setMaximumSize(QSize(85, 16777215))
        self.label_4.setFont(font)

        self.gridLayout.addWidget(self.label_4, 1, 0, 1, 1)

        self.widget_4 = QWidget(Protocol_Settings)
        self.widget_4.setObjectName(u"widget_4")

        self.gridLayout.addWidget(self.widget_4, 7, 0, 1, 1)

        self.label_2 = QLabel(Protocol_Settings)
        self.label_2.setObjectName(u"label_2")
        font2 = QFont()
        font2.setPointSize(9)
        font2.setBold(True)
        self.label_2.setFont(font2)

        self.gridLayout.addWidget(self.label_2, 8, 0, 1, 1)

        self.checkBox_live_variables = QCheckBox(Protocol_Settings)
        self.checkBox_live_variables.setObjectName(u"checkBox_live_variables")

        self.gridLayout.addWidget(self.checkBox_live_variables, 10, 0, 1, 6)

        self.checkBox_live_comments = QCheckBox(Protocol_Settings)
        self.checkBox_live_comments.setObjectName(u"checkBox_live_comments")

        self.gridLayout.addWidget(self.checkBox_live_comments, 11, 0, 1, 6)


        self.retranslateUi(Protocol_Settings)

        QMetaObject.connectSlotsByName(Protocol_Settings)
    # setupUi

    def retranslateUi(self, Protocol_Settings):
        Protocol_Settings.setWindowTitle(QCoreApplication.translate("Protocol_Settings", u"Form", None))
        self.pushButton_add_variable.setText(QCoreApplication.translate("Protocol_Settings", u"+", None))
        self.lineEdit_filename.setText(QCoreApplication.translate("Protocol_Settings", u"Datafile", None))
        self.lineEdit_filename.setPlaceholderText(QCoreApplication.translate("Protocol_Settings", u"Filename", None))
        self.label.setText(QCoreApplication.translate("Protocol_Settings", u"Filename:", None))
        self.pushButton_remove_variable.setText(QCoreApplication.translate("Protocol_Settings", u"-", None))
        self.radioButton_h5_after.setText(QCoreApplication.translate("Protocol_Settings", u"write hdf5 after run", None))
        self.checkBox_json_exp.setText(QCoreApplication.translate("Protocol_Settings", u"Export metadata to json", None))
        self.checkBox_no_config.setText(QCoreApplication.translate("Protocol_Settings", u"do not reconfigure instruments at start", None))
        self.checkBox_csv_exp.setText(QCoreApplication.translate("Protocol_Settings", u"Export data to csv", None))
        self.checkBox_NeXus.setText(QCoreApplication.translate("Protocol_Settings", u"also write NeXus entry to data", None))
        self.radioButton_h5_during.setText(QCoreApplication.translate("Protocol_Settings", u"write hdf5 during run", None))
        self.label_title.setText(QCoreApplication.translate("Protocol_Settings", u"General Configuration", None))
        self.label_4.setText(QCoreApplication.translate("Protocol_Settings", u"Protocol Name:", None))
        self.label_2.setText(QCoreApplication.translate("Protocol_Settings", u"Variables", None))
        self.checkBox_live_variables.setText(QCoreApplication.translate("Protocol_Settings", u"allow for live resetting of variables", None))
        self.checkBox_live_comments.setText(QCoreApplication.translate("Protocol_Settings", u"allow live comments to protocol", None))
    # retranslateUi




from PySide6.QtWidgets import QComboBox, QLabel, QCheckBox
from PySide6.QtGui import QFont

from nomad_camels.main_classes.loop_step import Loop_Step_Config
from nomad_camels.utility import variables_handling
from nomad_camels.ui_widgets.channels_check_table import Channels_Check_Table
from nomad_camels.utility.load_save_helper_functions import load_plots
from nomad_camels.bluesky_handling import builder_helper_functions
from nomad_camels.frontpanels.plot_definer import Plot_Button_Overview

from nomad_camels.loop_steps.for_while_loops import (
    For_Loop_Step_Config_Sub,
    For_Loop_Step,
)


class Simple_Sweep(For_Loop_Step):
    """
    A sweep over a single channel reading arbitrary other channels. The
    information about which values to sweep through are inherited from
    For_Loop_Step.

    Attributes
    ----------
    sweep_channel : str
        The name of the channel that should be sweeped.
    data_output : str
        Whether the data is put into its own stream('sub-stream') or the primary
        stream ('main-stream').
    plots : list[Plot_Info]
        List of the plots that should be created for the sweep.
    read_channels : list[str]
        List of the channels that should be read during the sweep.
    """

    def __init__(
        self, name="", children=None, parent_step=None, step_info=None, **kwargs
    ):
        super().__init__(name, children, parent_step, step_info, **kwargs)
        step_info = step_info or {}
        self.step_type = "Simple Sweep"
        self.has_children = False
        self.sweep_channel = (
            step_info["sweep_channel"] if "sweep_channel" in step_info else ""
        )
        self.data_output = (
            step_info["data_output"] if "data_output" in step_info else "sub-stream"
        )
        self.plots = load_plots([], step_info["plots"]) if "plots" in step_info else []
        self.read_channels = (
            step_info["read_channels"] if "read_channels" in step_info else []
        )

    def update_used_devices(self):
        """Includes the devices from the read_channels and the sweep_channel."""
        self.used_devices = []
        set_device = variables_handling.channels[self.sweep_channel].device
        self.used_devices.append(set_device)
        for channel in self.read_channels:
            if channel in variables_handling.channels:
                device = variables_handling.channels[channel].device
                if device not in self.used_devices:
                    self.used_devices.append(device)

    def update_variables(self):
        """Adds the fit variables from the plots."""
        variables = {}
        stream = f"{self.name}"
        for plot in self.plots:
            variables.update(plot.get_fit_vars(stream))
        variables_handling.loop_step_variables.update(variables)
        super().update_variables()

    def get_outer_string(self):
        """Gives the string to create the plots of the sweeps"""
        if self.plots:
            return builder_helper_functions.plot_creator(
                self.plots, f"create_plots_{self.name}"
            )[0]
        return ""

    def get_add_main_string(self):
        """Calling the plot_creator from steps_add_main"""
        stream = f'"{self.name}"'
        if self.data_output == "main stream":
            stream = '"primary"'
        add_main_string = ""
        if self.plots:
            add_main_string += builder_helper_functions.get_plot_add_string(
                self.name, stream
            )
        return add_main_string

    def get_protocol_string(self, n_tabs=1):
        """The channels to be read are set up. Then a for loop for the sweep is
        started, setting the channel, then reading."""
        tabs = "\t" * n_tabs

        stream = f'"{self.name}"'
        if self.data_output == "main stream":
            stream = "stream_name"

        protocol_string = f"{tabs}channels = ["
        for i, channel in enumerate(self.read_channels):
            if channel not in variables_handling.channels:
                raise Exception(
                    f"Trying to read channel {channel} in {self.full_name}, but it does not exist!"
                )
            if i > 0:
                protocol_string += ", "
            name = variables_handling.channels[channel].name
            if "." in name:
                dev, chan = name.split(".")
                protocol_string += f'devs["{dev}"].{chan}'
            else:
                protocol_string += f'devs["{name}"]'
        protocol_string += "]\n"
        protocol_string += f"{tabs}helper_functions.clear_plots(plots, {stream})"
        protocol_string += super().get_protocol_string(n_tabs)
        name = variables_handling.channels[self.sweep_channel].name
        if "." in name:
            dev, chan = name.split(".")
            setter = f'devs["{dev}"].{chan}'
        else:
            setter = f'devs["{name}"]'

        protocol_string += f'{tabs}\tyield from bps.abs_set({setter}, {self.name.replace(" ", "_")}_Value, group="A")\n'
        protocol_string += f'{tabs}\tyield from bps.wait("A")\n'
        protocol_string += (
            f"{tabs}\tyield from bps.trigger_and_read(channels, name={stream})\n"
        )
        protocol_string += f"{tabs}yield from helper_functions.get_fit_results(all_fits, namespace, True, {stream})\n"
        self.update_time_weight()
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """The read and sweep channels and the sweep values are specified."""
        vals = super().get_protocol_short_string(n_tabs)[:-1].split(" for ")[-1]
        tabs = "\t" * n_tabs
        short_string = f"{tabs}Sweep '{self.name}' {self.sweep_channel}:\n{tabs}\tRead: {self.read_channels}\n{tabs}\tValues: {vals}"
        return short_string


class Simple_Sweep_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Simple_Sweep, parent=None):
        super().__init__(parent, loop_step)
        self.loop_step = loop_step
        label_sweep_channel = QLabel("Sweep Channel:")
        out_box = []
        in_box = []
        for channel in variables_handling.channels:
            in_box.append(channel)
            if variables_handling.channels[channel].output:
                out_box.append(channel)
        self.comboBox_sweep_channel = QComboBox()
        self.comboBox_sweep_channel.addItems(out_box)
        if loop_step.sweep_channel in out_box:
            self.comboBox_sweep_channel.setCurrentText(loop_step.sweep_channel)

        label_data = QLabel("Data Output:")
        self.comboBox_data_output = QComboBox()
        output_types = ["sub-stream", "main stream", "own file"]
        self.comboBox_data_output.addItems(output_types)
        self.comboBox_data_output.setCurrentText(loop_step.data_output)

        self.sweep_widget = For_Loop_Step_Config_Sub(parent=self, loop_step=loop_step)

        # self.read_table = AddRemoveTable(title='Read Channels', headerLabels=[],
        #                                  tableData=loop_step.read_channels,
        #                                  comboBoxes=in_box)
        labels = ["read", "channel"]
        info_dict = {"channel": self.loop_step.read_channels}
        self.read_table = Channels_Check_Table(
            self, labels, info_dict=info_dict, title="Read-Channels"
        )

        # self.checkBox_use_own_plots = QCheckBox('Use own Plots')
        # self.checkBox_use_own_plots.setChecked(loop_step.use_own_plots)

        self.plot_widge = Plot_Button_Overview(self, self.loop_step.plots)

        # label_proc = QLabel('Data processing')
        font = QFont()
        font.setBold(True)
        # label_proc.setStyleSheet('font-size: 9pt')
        # label_proc.setFont(font)
        # self.checkBox_minmax = QCheckBox('Calculate min/max')
        # self.checkBox_minmax.setChecked(loop_step.calc_minmax)
        # self.checkBox_mean = QCheckBox('Calculate mean')
        # self.checkBox_mean.setChecked(loop_step.calc_mean)
        # self.checkBox_stddev = QCheckBox('Calculate standard deviation')
        # self.checkBox_stddev.setChecked(loop_step.calc_stddev)
        # self.checkBox_fit = QCheckBox('Calculate fit')
        #         # self.checkBox_fit.setChecked(loop_step.calc_fit)
        # self.checkBox_fit.clicked.connect(self.change_fitting)
        # self.checkBox_guess_fit = QCheckBox('Guess initial params')
        # self.checkBox_guess_fit.setChecked(loop_step.guess_fit_params)
        # self.checkBox_guess_fit.clicked.connect(self.change_fitting)
        # self.radioButton_predev = QRadioButton('Predefined function')
        # self.radioButton_own = QRadioButton('Own function')
        # self.radioButton_predev.setChecked(True)
        # self.radioButton_own.setChecked(loop_step.use_custom_fit)
        # self.radioButton_predev.clicked.connect(self.change_fitting)
        # self.radioButton_own.clicked.connect(self.change_fitting)
        #
        # self.comboBox_fit = QComboBox()
        # self.comboBox_fit.addItems(sorted(models_names.keys()))
        # if loop_step.predef_fit in models_names.keys():
        #     self.comboBox_fit.setCurrentText(loop_step.predef_fit)
        # else:
        #     self.comboBox_fit.setCurrentText('Linear')
        # self.lineEdit_fit_func = QLineEdit()
        # self.lineEdit_fit_func.setText(loop_step.custom_fit)
        # self.comboBox_fit.currentTextChanged.connect(self.change_fitting)
        # self.lineEdit_fit_func.textChanged.connect(self.change_fitting)
        #
        # cols = ['name', 'initial value']
        # self.start_params = AddRemoveTable(headerLabels=cols,
        #                                   title='Fit Parameters',
        #                                   editables=[1],
        #                                   tableData=loop_step.fit_params)
        # self.start_params.addButton.setHidden(True)
        # self.start_params.removeButton.setHidden(True)
        # self.change_fitting()

        self.layout().addWidget(label_sweep_channel, 1, 0)
        self.layout().addWidget(self.comboBox_sweep_channel, 1, 1, 1, 4)
        self.layout().addWidget(label_data, 2, 0)
        self.layout().addWidget(self.comboBox_data_output, 2, 1, 1, 4)
        self.layout().addWidget(self.sweep_widget, 5, 0, 1, 5)
        self.layout().addWidget(self.read_table, 6, 0, 1, 5)

        self.layout().addWidget(self.plot_widge, 8, 0, 1, 5)
        # self.layout().addWidget(self.checkBox_use_own_plots, 7, 0, 1, 5)
        # self.layout().addWidget(self.plot_table, 7, 2, 1, 3)
        # self.checkBox_use_own_plots.clicked.connect(self.use_plot_change)

        # self.layout().addWidget(label_proc, 10, 0, 1, 5)
        # self.layout().addWidget(self.checkBox_minmax, 11, 0, 1, 2)
        # self.layout().addWidget(self.checkBox_mean, 11, 2, 1, 3)
        # self.layout().addWidget(self.checkBox_stddev, 13, 0, 1, 2)
        # self.layout().addWidget(self.checkBox_fit, 20, 0, 1, 2)
        # self.layout().addWidget(self.checkBox_guess_fit, 20, 2, 1, 3)
        # self.layout().addWidget(self.radioButton_predev, 21, 0, 1, 2)
        # self.layout().addWidget(self.radioButton_own, 21, 2, 1, 3)
        # self.layout().addWidget(self.comboBox_fit, 22, 0, 1, 2)
        # self.layout().addWidget(self.lineEdit_fit_func, 22, 2, 1, 3)
        # self.layout().addWidget(self.start_params, 23, 0, 1, 5)

        # self.use_plot_change()

    # def setup_plots(self):
    #     """Called when any preferences are changed. Makes the dictionary
    #      of preferences and calls save_preferences from the
    #      load_save_functions module."""
    #     plot_dialog = Plot_Definer(self)
    #     plot_dialog.exec()
    #     print(plot_dialog.data)
    # if settings_dialog.exec():
    #     self.preferences = settings_dialog.get_settings()
    #     number_formatting.preferences = self.preferences
    #     self.toggle_dark_mode()
    #     load_save_functions.save_preferences(self.preferences)
    #     variables_handling.device_driver_path = self.preferences['device_driver_path']
    #     variables_handling.meas_files_path = self.preferences['meas_files_path']
    # prefs = {'autosave': self.actionAutosave_on_closing.isChecked(),
    #          'dark_mode': self.actionDark_Mode.isChecked()}
    # load_save_functions.save_preferences(prefs)

    # def use_plot_change(self):
    #     """ """
    #     use_plots = self.checkBox_use_own_plots.isChecked()
    #     self.plot_widge.setEnabled(use_plots)

    def update_step_config(self):
        """ """
        super().update_step_config()
        # self.loop_step.use_own_plots = self.checkBox_use_own_plots.isChecked()
        self.loop_step.plots = self.plot_widge.plot_data
        # self.loop_step.plots = self.plot_table.update_table_data()
        self.loop_step.read_channels = self.read_table.get_info()["channel"]
        self.loop_step.data_output = self.comboBox_data_output.currentText()
        self.loop_step.sweep_channel = self.comboBox_sweep_channel.currentText()
        # self.loop_step.calc_minmax = self.checkBox_minmax.isChecked()
        # self.loop_step.calc_mean = self.checkBox_mean.isChecked()
        # self.loop_step.calc_stddev = self.checkBox_stddev.isChecked()
        # # self.loop_step.calc_fit = self.checkBox_fit.isChecked()
        # self.loop_step.use_custom_fit = self.radioButton_own.isChecked()
        # self.loop_step.predef_fit = self.comboBox_fit.currentText()
        # self.loop_step.custom_fit = self.lineEdit_fit_func.text()



import time
import numpy as np
from PySide6.QtWidgets import QCheckBox, QComboBox, QLabel, QWidget, QGridLayout, QStyle
from PySide6.QtCore import QThread, Signal, Qt
from PySide6.QtGui import QKeyEvent

from nomad_camels.main_classes.manual_control import (
    Manual_Control,
    Manual_Control_Config,
)
from nomad_camels.utility import variables_handling, device_handling, number_formatting

from .ui_stage_control import Ui_Form


class Stage_Control(Manual_Control, Ui_Form):
    """
    A class that provides a GUI for controlling a stage.

    Parameters
    ----------
    parent : QWidget
        The parent widget.
    control_data : dict
        A dictionary containing the control data.
        The following keys are expected:
        - name: The name of the control.
        - use_axis: A list of three booleans, indicating which axes are used.
        - axis_channel: A list of three strings, indicating the channels used for each axis.
        - read_axis: A list of three booleans, indicating which axes are read back.
        - read_channel: A list of three strings, indicating the channels used for reading back. If an axis is not read back, the corresponding entry should be None.
        - axis_ref: A list of three functions, indicating the reference drive functions for each axis. If an axis is does not have a reference, the corresponding entry should be None.
        - axis_stop: A list of three functions, indicating the stop functions for each axis. If an axis does not have a stop function, the corresponding entry should be None.
        - auto_reference: A list of three booleans, indicating whether the axes should automatically perform their reference function at start.
        - manual_function: A list of three functions, indicating the manual move functions for each axis. If an axis does not have a manual move function, the corresponding entry should be None.
        - read_frequ: A float, indicating the frequency at which the axes are read back.
        - manual_active: A boolean, indicating whether manual control is active by default.
        - stepSize_X: A float, indicating the step size for the X axis.
        - stepSize_Y: A float, indicating the step size for the Y axis.
        - stepSize_Z: A float, indicating the step size for the Z axis.
        - manual_X: A float, indicating the manual move speed for the X axis.
        - manual_Y: A float, indicating the manual move speed for the Y axis.
        - manual_Z: A float, indicating the manual move speed for the Z axis.
    """

    def __init__(self, parent=None, control_data=None):
        control_data = control_data or {}
        if "name" in control_data:
            name = control_data["name"]
        else:
            name = "Stage Control"
        super().__init__(parent=parent, title=name)
        self.setupUi(self)
        self.setWindowTitle(f"{name} - NOMAD CAMELS")
        self.control_data = control_data

        self.pushButton_up.setIcon(self.style().standardIcon(QStyle.SP_ArrowUp))
        self.pushButton_down.setIcon(self.style().standardIcon(QStyle.SP_ArrowDown))
        self.pushButton_left.setIcon(self.style().standardIcon(QStyle.SP_ArrowLeft))
        self.pushButton_right.setIcon(self.style().standardIcon(QStyle.SP_ArrowRight))
        self.pushButton_zUp.setIcon(self.style().standardIcon(QStyle.SP_ArrowUp))
        self.pushButton_zDown.setIcon(self.style().standardIcon(QStyle.SP_ArrowDown))
        self.pushButton_position.setIcon(
            self.style().standardIcon(QStyle.SP_MediaSeekBackward)
        )
        self.pushButton_stop.setIcon(self.style().standardIcon(QStyle.SP_BrowserStop))

        use_x, use_y, use_z = control_data["use_axis"]
        self.pushButton_left.setEnabled(use_x)
        self.pushButton_right.setEnabled(use_x)
        self.pushButton_up.setEnabled(use_y)
        self.pushButton_down.setEnabled(use_y)
        self.pushButton_zUp.setEnabled(use_z)
        self.pushButton_zDown.setEnabled(use_z)

        self.pushButton_right.clicked.connect(
            lambda state=None, axis=0: self.step_axis(axis)
        )
        self.pushButton_left.clicked.connect(
            lambda state=None, axis=0, u=False: self.step_axis(axis, u)
        )
        self.pushButton_up.clicked.connect(
            lambda state=None, axis=1: self.step_axis(axis)
        )
        self.pushButton_down.clicked.connect(
            lambda state=None, axis=1, u=False: self.step_axis(axis, u)
        )
        self.pushButton_zUp.clicked.connect(
            lambda state=None, axis=2: self.step_axis(axis)
        )
        self.pushButton_zDown.clicked.connect(
            lambda state=None, axis=2, u=False: self.step_axis(axis, u)
        )

        self.pushButton_ref.clicked.connect(self.reference_drive)
        self.pushButton_stop.clicked.connect(self.stop_moving)

        self.pushButton_position.clicked.connect(self.input_position)
        self.pushButton_go_to.clicked.connect(self.move_to_position)

        self.lines = [
            self.lineEdit_read_frequ,
            self.lineEdit_stepX,
            self.lineEdit_stepY,
            self.lineEdit_stepZ,
            self.lineEdit_manualX,
            self.lineEdit_manualY,
            self.lineEdit_manualZ,
            self.lineEdit_goX,
            self.lineEdit_goY,
            self.lineEdit_goZ,
        ]
        self.line_names = [
            "read_frequ",
            "stepSize_X",
            "stepSize_Y",
            "stepSize_Z",
            "manual_X",
            "manual_Y",
            "manual_Z",
            "go_to_X",
            "go_to_Y",
            "go_to_Z",
        ]
        self.checks = [
            self.checkBox_refX,
            self.checkBox_refY,
            self.checkBox_refZ,
            self.checkBox_manualActive,
        ]
        self.check_names = ["find_ref_X", "find_ref_Y", "find_ref_Z", "manual_active"]
        if "read_frequ" not in control_data:
            control_data["read_frequ"] = 0.5
        if "manual_active" not in control_data:
            control_data["manual_active"] = False

        for i, line in enumerate(self.lines):
            if self.line_names[i] not in control_data:
                control_data[self.line_names[i]] = 0
            line.setText(str(control_data[self.line_names[i]]))
            line.textChanged.connect(self.line_change)
        for i, check in enumerate(self.checks):
            if self.check_names[i] not in control_data:
                control_data[self.check_names[i]] = True
            check.setChecked(control_data[self.check_names[i]])
            check.clicked.connect(self.check_change)

        ax_names = ["X", "Y", "Z"]
        set_channels = []
        read_channels = []
        ref_functions = []
        stop_functions = []
        manual_functions = []
        for i, use in enumerate(control_data["use_axis"]):
            ax = ax_names[i]
            for j, name in enumerate(self.line_names):
                if ax in name:
                    self.lines[j].setEnabled(use)
            for j, name in enumerate(self.check_names):
                if ax in name:
                    self.checks[j].setEnabled(use)
            if not use:
                set_channels.append("None")
                read_channels.append("None")
                ref_functions.append("None")
                stop_functions.append("None")
                manual_functions.append("None")
                continue
            set_channels.append(control_data["axis_channel"][i])
            if control_data["read_axis"][i]:
                read_channels.append(control_data["read_channel"][i])
            else:
                read_channels.append("None")
            ref_functions.append(control_data["axis_ref"][i])
            stop_functions.append(control_data["axis_stop"][i])
            manual_functions.append(control_data["manual_function"][i])

        channels = set(read_channels + set_channels)
        if "None" in channels:
            channels.remove("None")
        channels = list(channels)
        self.set_channels = set_channels
        self.read_channels = read_channels
        self.ref_funcs = ref_functions
        self.stop_funcs = stop_functions
        self.manual_funcs = manual_functions
        self.read_thread = None
        self.move_thread = None
        self.start_multiple_devices(channels, True)

    def device_ready(self):
        """
        Overwrites the `device_ready` method from the `Manual_Control` class. This method is called when the device is ready. It sets the channels, read channels, reference functions, stop functions, and manual functions from the instantiated devices. It also starts the readback thread and the move thread.
        """
        super().device_ready()
        self.set_channels = device_handling.get_channels_from_string_list(
            self.set_channels
        )
        self.read_channels = device_handling.get_channels_from_string_list(
            self.read_channels
        )
        self.ref_funcs = device_handling.get_functions_from_string_list(self.ref_funcs)
        self.stop_funcs = device_handling.get_functions_from_string_list(
            self.stop_funcs
        )
        self.manual_funcs = device_handling.get_functions_from_string_list(
            self.manual_funcs
        )

        read_not_none = False
        for channel in self.read_channels:
            if channel is not None:
                read_not_none = True
                break
        if read_not_none:
            self.start_read_thread()
        else:
            self.lineEdit_read_frequ.setEnabled(False)
        self.start_move_thread()
        for child in self.children():
            if isinstance(child, QWidget):
                child.setFocusPolicy(Qt.ClickFocus)
        self.setFocusPolicy(Qt.ClickFocus)
        self.show()
        for i, auto in enumerate(self.control_data["auto_reference"]):
            self.checks[i].setChecked(auto)
        self.reference_drive()
        for i, auto in enumerate(self.control_data["auto_reference"]):
            self.checks[i].setChecked(True)

    def start_move_thread(self):
        read_not_none = False
        for channel in self.read_channels:
            if channel is not None:
                read_not_none = True
                break
        positions = [np.nan, np.nan, np.nan]
        if read_not_none:
            for i in range(3):
                try:
                    positions[i] = self.read_channels[i].get()
                except:
                    pass
        manual_X = self.control_data["manual_X"]
        manual_Y = self.control_data["manual_Y"]
        manual_Z = self.control_data["manual_Z"]
        speeds = [manual_X, manual_Y, manual_Z]
        self.move_thread = Move_Thread(
            self,
            self.set_channels,
            speeds,
            manual_functions=self.manual_funcs,
            stop_functions=self.stop_funcs,
            starting_positions=positions,
        )
        self.move_thread.exception_signal.connect(
            lambda ex, name="move": self.exception_caught(ex, name)
        )
        self.move_thread.start()

    def start_read_thread(self):
        self.read_thread = Readback_Thread(
            self, self.read_channels, self.control_data["read_frequ"]
        )
        self.read_thread.data_sig.connect(self.update_readback)
        self.read_thread.exception_signal.connect(
            lambda ex, name="read": self.exception_caught(ex, name)
        )
        self.read_thread.start()

    def exception_caught(self, ex, name):
        """
        Handles exceptions caught by the readback or move threads.

        Parameters
        ----------
        ex : Exception
            The exception that was caught.

        name : str
            The name of the thread that caught the exception.
        """
        from PySide6.QtWidgets import QMessageBox

        msg = QMessageBox()
        msg.setIcon(QMessageBox.Critical)
        msg.setText("Error in read thread")
        msg.setInformativeText(
            f"An error occured in the {name} thread of {self.name}:\n{ex}\nDo you want to restart the read thread?"
        )
        msg.setWindowTitle("Error in read thread")
        msg.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        msg.setDefaultButton(QMessageBox.Yes)
        if msg.exec() == QMessageBox.Yes:
            if name == "read":
                self.start_read_thread()
            elif name == "move":
                self.start_move_thread()

    def line_change(self):
        """
        Updates the control data when a line edit is changed.
        This method is connected to the `textChanged` signal of the line edits, including the read frequency, step sizes, manual speeds, and go-to positions.
        """
        for i, line in enumerate(self.lines):
            try:
                self.control_data[self.line_names[i]] = float(line.text())
            except:
                pass
        if self.read_thread:
            self.read_thread.read_time = self.control_data["read_frequ"]
        manual_X = self.control_data["manual_X"]
        manual_Y = self.control_data["manual_Y"]
        manual_Z = self.control_data["manual_Z"]
        self.move_thread.move_speeds = [manual_X, manual_Y, manual_Z]

    def check_change(self):
        """
        Updates the control data when a check box is changed. This method is connected to the `clicked` signal of the check boxes, including the reference checks and the manual active check.
        """
        for i, check in enumerate(self.checks):
            self.control_data[self.check_names[i]] = check.isChecked()

    def update_readback(self, x, y, z):
        """
        Updates the readback values in the GUI. This method is connected to the `data_sig` signal of the readback thread.

        Parameters
        ----------
        x : float
            The readback value of the X axis.

        y : float
            The readback value of the Y axis.

        z : float
            The readback value of the Z axis.
        """
        self.lineEdit_currentX.setText(number_formatting.format_number(x))
        self.lineEdit_currentY.setText(number_formatting.format_number(y))
        self.lineEdit_currentZ.setText(number_formatting.format_number(z))

    def close(self) -> bool:
        """
        Stops the readback and move threads when the control is closed."""
        if self.read_thread:
            self.read_thread.still_running = False
        if self.move_thread:
            self.move_thread.still_running = False
        return super().close()

    def closeEvent(self, a0) -> None:
        """
        Stops the readback and move threads when the control is closed.

        Parameters
        ----------
        a0 : QCloseEvent
        """
        if self.read_thread:
            self.read_thread.still_running = False
        if self.move_thread:
            self.move_thread.still_running = False
        return super().closeEvent(a0)

    def step_axis(self, axis, up=True):
        """

        Parameters
        ----------
        axis :

        up :
             (Default value = True)

        Returns
        -------

        """
        if self.read_thread:
            self.read_thread.paused = True
        ax_names = ["X", "Y", "Z"]
        step_size = self.control_data[f"stepSize_{ax_names[axis]}"]
        if not up:
            step_size *= -1
        move_to = [np.nan] * 3
        for i in range(3):
            if self.read_channels[i] is not None:
                move_to[i] = self.read_channels[i].get()
            elif self.set_channels[i] is not None:
                move_to[i] = self.set_channels[i].get()
        if self.read_thread:
            self.read_thread.paused = False
        move_to[axis] = move_to[axis] + step_size
        self.move_thread.set_absolute = move_to

    def reference_drive(self):
        """ """
        self.setCursor(Qt.WaitCursor)
        try:
            checks = [self.checkBox_refX, self.checkBox_refY, self.checkBox_refZ]
            for i, func in enumerate(self.ref_funcs):
                if checks[i].isChecked() and func:
                    func()
        finally:
            self.setCursor(Qt.ArrowCursor)

    def stop_moving(self):
        """ """
        for func in self.stop_funcs:
            if func:
                func()

    def input_position(self):
        """ """
        self.lineEdit_goX.setText(self.lineEdit_currentX.text())
        self.lineEdit_goY.setText(self.lineEdit_currentY.text())
        self.lineEdit_goZ.setText(self.lineEdit_currentZ.text())

    def move_to_position(self):
        """ """
        axes = ["X", "Y", "Z"]
        positions = [np.nan, np.nan, np.nan]
        for i in range(3):
            if self.set_channels[i]:
                positions[i] = self.control_data[f"go_to_{axes[i]}"]
        self.move_thread.set_absolute = positions

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        """

        Parameters
        ----------
        a0: QKeyEvent :


        Returns
        -------

        """
        if a0.isAutoRepeat():
            return super().keyPressEvent(a0)
        if (
            not self.checkBox_manualActive.isChecked()
            or a0.modifiers() != Qt.ControlModifier
        ):
            return super().keyPressEvent(a0)
        if a0.key() == Qt.Key_Left and (self.set_channels[0] or self.manual_funcs[0]):
            self.move_thread.up_dir[0] = False
            self.move_thread.movers[0] = True
        elif a0.key() == Qt.Key_Right and (
            self.set_channels[0] or self.manual_funcs[0]
        ):
            self.move_thread.up_dir[0] = True
            self.move_thread.movers[0] = True
        elif a0.key() == Qt.Key_Up and (self.set_channels[1] or self.manual_funcs[1]):
            self.move_thread.up_dir[1] = True
            self.move_thread.movers[1] = True
        elif a0.key() == Qt.Key_Down and (self.set_channels[1] or self.manual_funcs[1]):
            self.move_thread.up_dir[1] = False
            self.move_thread.movers[1] = True
        elif a0.key() == Qt.Key_PageDown and (
            self.set_channels[2] or self.manual_funcs[2]
        ):
            self.move_thread.up_dir[2] = False
            self.move_thread.movers[2] = True
        elif a0.key() == Qt.Key_PageUp and (
            self.set_channels[2] or self.manual_funcs[2]
        ):
            self.move_thread.up_dir[2] = True
            self.move_thread.movers[2] = True
        else:
            super().keyPressEvent(a0)

    def keyReleaseEvent(self, a0: QKeyEvent) -> None:
        """

        Parameters
        ----------
        a0: QKeyEvent :


        Returns
        -------

        """
        if a0.isAutoRepeat():
            return super().keyPressEvent(a0)
        if a0.key() == Qt.Key_Left:
            self.move_thread.movers[0] = False
        elif a0.key() == Qt.Key_Right:
            self.move_thread.movers[0] = False
        elif a0.key() == Qt.Key_Up:
            self.move_thread.movers[1] = False
        elif a0.key() == Qt.Key_Down:
            self.move_thread.movers[1] = False
        elif a0.key() == Qt.Key_PageDown:
            self.move_thread.movers[2] = False
        elif a0.key() == Qt.Key_PageUp:
            self.move_thread.movers[2] = False
        else:
            super().keyReleaseEvent(a0)


class Move_Thread(QThread):
    """ """

    exception_signal = Signal(Exception)

    def __init__(
        self,
        parent=None,
        channels=None,
        move_speeds=None,
        manual_functions=None,
        stop_functions=None,
        starting_positions=None,
    ):
        super().__init__(parent=parent)
        self.channels = channels or []
        self.manual_functions = manual_functions or []
        self.stop_functions = stop_functions or []
        self.still_running = True
        self.movers = [False, False, False]
        self.move_speeds = move_speeds or [0, 0, 0]
        self.move_starts = [np.nan, np.nan, np.nan]
        self.up_dir = [True, True, True]
        self.last_set = starting_positions or [np.nan, np.nan, np.nan]
        self.moving_started = [False, False, False]
        self.set_absolute = [np.nan, np.nan, np.nan]
        self.last_absolute = [np.nan, np.nan, np.nan]

    def run(self) -> None:
        """ """
        try:
            while self.still_running:
                move = False
                for i, mover in enumerate(self.movers):
                    if mover:
                        move = True
                        if np.isnan(self.move_starts[i]):
                            self.move_starts[i] = time.time()
                            self.last_set[i] = self.channels[i].get()
                        self.move(i)
                    else:
                        if not np.isnan(self.move_starts[i]):
                            self.move_starts[i] = np.nan
                            if self.stop_functions and self.stop_functions[i]:
                                self.stop_functions[i]()
                            self.moving_started[i] = False
                if self.set_absolute != self.last_absolute:
                    for i, val in enumerate(self.set_absolute):
                        if not np.isnan(val):
                            self.channels[i].put(val)
                            self.last_absolute[i] = val
                    continue
                if not move:
                    time.sleep(0.1)
        except Exception as e:
            self.exception_signal.emit(e)

    def move(self, ax):
        """

        Parameters
        ----------
        ax :


        Returns
        -------

        """
        if self.manual_functions and self.manual_functions[ax]:
            if not self.moving_started[ax]:
                self.manual_functions[ax](
                    self.move_speeds[ax] * (1 if self.up_dir[ax] else -1)
                )
        else:
            before = self.last_set[ax]
            now = time.time()
            step_size = self.move_speeds[ax] * (now - self.move_starts[ax])
            self.move_starts[ax] = now
            if not self.up_dir[ax]:
                step_size *= -1
            val = step_size + before
            self.channels[ax].put(val)
            self.last_set[ax] = val
        self.moving_started[ax] = True


class Readback_Thread(QThread):
    """ """

    data_sig = Signal(float, float, float)
    exception_signal = Signal(Exception)

    def __init__(self, parent=None, channels=None, read_time=np.inf):
        super().__init__(parent=parent)
        self.channels = channels or []
        self.read_time = read_time
        self.still_running = True
        self.paused = False

    def run(self):
        """ """
        try:
            accum = 0
            while self.still_running:
                if self.paused:
                    time.sleep(self.read_time)
                    continue
                self.do_reading()
                if self.read_time > 5:
                    if self.read_time - accum > 5:
                        time.sleep(5)
                        accum += 5
                        continue
                    else:
                        time.sleep(self.read_time - accum)
                        accum = 0
                else:
                    time.sleep(self.read_time)
        except Exception as e:
            self.exception_signal.emit(e)

    def do_reading(self):
        """ """
        vals = []
        for channel in self.channels:
            if channel:
                vals.append(channel.get())
            else:
                vals.append(np.nan)
        self.data_sig.emit(*vals)


class Stage_Control_Config(Manual_Control_Config):
    """ """

    def __init__(self, parent=None, control_data=None):
        super().__init__(
            parent=parent,
            control_data=control_data,
            title="Stage Control Config",
            control_type="Stage_Control",
        )
        control_data = control_data or {}
        self.axis_checkboxes = []
        self.channels_combos = []
        self.read_checkboxes = []
        self.read_combos = []
        self.ref_combos = []
        # self.ref_vals = []
        self.stop_combos = []
        # self.stop_vals = []
        self.auto_ref_checkboxes = []
        self.manual_move_combos = []
        outputs = variables_handling.get_output_channels()
        functions = variables_handling.get_non_channel_functions()
        channels = list(variables_handling.channels.keys())
        help_widge = QWidget()
        layout = QGridLayout()
        help_widge.setLayout(layout)
        for i, ax in enumerate(["x", "y", "z"]):
            # label = QLabel(ax)
            # layout.addWidget(label, 2+i, 0)

            axis_box = QCheckBox(f"use axis {ax}")
            if "use_axis" in control_data and control_data["use_axis"]:
                axis_box.setChecked(control_data["use_axis"][i])
            self.axis_checkboxes.append(axis_box)
            layout.addWidget(axis_box, 2 + i, 0)

            channel_combo = QComboBox()
            channel_combo.addItems(outputs)
            if (
                "axis_channel" in control_data
                and control_data["axis_channel"][i] in outputs
            ):
                channel_combo.setCurrentText(control_data["axis_channel"][i])
            self.channels_combos.append(channel_combo)
            layout.addWidget(channel_combo, 2 + i, 1)

            read_box = QCheckBox(f"readback axis {ax}")
            self.read_checkboxes.append(read_box)
            if "read_axis" in control_data and control_data["read_axis"]:
                read_box.setChecked(control_data["read_axis"][i])
            layout.addWidget(read_box, 2 + i, 2)

            read_combo = QComboBox()
            read_combo.addItems(channels)
            if (
                "read_channel" in control_data
                and control_data["read_channel"][i] in channels
            ):
                read_combo.setCurrentText(control_data["read_channel"][i])
            self.read_combos.append(read_combo)
            layout.addWidget(read_combo, 2 + i, 3)

            label = QLabel(f"manual move function {ax}:")
            layout.addWidget(label, 2 + i, 4)

            manual_combo = QComboBox()
            manual_combo.addItems(functions + ["None"])
            if (
                "manual_function" in control_data
                and control_data["manual_function"][i] in functions
            ):
                manual_combo.setCurrentText(control_data["manual_function"][i])
            else:
                manual_combo.setCurrentText("None")
            self.manual_move_combos.append(manual_combo)
            layout.addWidget(manual_combo, 2 + i, 5)

            label = QLabel(f"reference function {ax}:")
            layout.addWidget(label, 10 + i, 0)

            ref_combo = QComboBox()
            ref_combo.addItems(functions + ["None"])
            if "axis_ref" in control_data and control_data["axis_ref"][i] in functions:
                ref_combo.setCurrentText(control_data["axis_ref"][i])
            else:
                ref_combo.setCurrentText("None")
            self.ref_combos.append(ref_combo)
            layout.addWidget(ref_combo, 10 + i, 1)

            # ref_val = QLineEdit()
            # if 'ref_vals' in control_data and control_data['ref_vals']:
            #     ref_val.setText(control_data['ref_vals'][i])
            # self.ref_vals.append(ref_val)
            # layout.addWidget(ref_val, 10+i, 2)

            label = QLabel(f"stop function {ax}:")
            layout.addWidget(label, 10 + i, 2)

            stop_combo = QComboBox()
            stop_combo.addItems(functions + ["None"])
            if (
                "axis_stop" in control_data
                and control_data["axis_stop"][i] in functions
            ):
                stop_combo.setCurrentText(control_data["axis_stop"][i])
            else:
                stop_combo.setCurrentText("None")
            self.stop_combos.append(stop_combo)
            layout.addWidget(stop_combo, 10 + i, 3)

            auto_ref = QCheckBox("auto-reference at start")
            if "auto_reference" in control_data:
                auto_ref.setChecked(control_data["auto_reference"][i])
            self.auto_ref_checkboxes.append(auto_ref)
            layout.addWidget(auto_ref, 10 + i, 4, 1, 2)

            # stop_val = QLineEdit()
            # if 'stop_vals' in control_data and control_data['stop_vals']:
            #     stop_val.setText(control_data['stop_vals'][i])
            # self.stop_vals.append(stop_val)
            # layout.addWidget(stop_val, 10+i, 5)

            axis_box.clicked.connect(self.change_usage)
            read_box.clicked.connect(self.change_usage)
        self.layout().addWidget(help_widge, 1, 0, 1, 2)
        self.change_usage()

    def change_usage(self):
        """ """
        for i, box in enumerate(self.axis_checkboxes):
            able = box.isChecked()
            readback = self.read_checkboxes[i].isChecked()
            self.channels_combos[i].setEnabled(able)
            self.read_checkboxes[i].setEnabled(able)
            self.read_combos[i].setEnabled(able and readback)
            self.ref_combos[i].setEnabled(able)
            # self.ref_vals[i].setEnabled(able)
            self.stop_combos[i].setEnabled(able)
            self.auto_ref_checkboxes[i].setEnabled(able)
            self.manual_move_combos[i].setEnabled(able)
            # self.stop_vals[i].setEnabled(able)

    def accept(self):
        """ """
        self.control_data["use_axis"] = []
        self.control_data["axis_channel"] = []
        self.control_data["read_axis"] = []
        self.control_data["read_channel"] = []
        self.control_data["axis_ref"] = []
        # self.control_data['ref_vals'] = []
        self.control_data["axis_stop"] = []
        # self.control_data['stop_vals'] = []
        self.control_data["auto_reference"] = []
        self.control_data["manual_function"] = []
        for i in range(3):
            self.control_data["use_axis"].append(self.axis_checkboxes[i].isChecked())
            self.control_data["axis_channel"].append(
                self.channels_combos[i].currentText()
            )
            self.control_data["read_axis"].append(self.read_checkboxes[i].isChecked())
            self.control_data["read_channel"].append(self.read_combos[i].currentText())
            # self.control_data['ref_vals'].append(self.ref_vals[i].text())
            self.control_data["axis_ref"].append(self.ref_combos[i].currentText())
            # self.control_data['stop_vals'].append(self.stop_vals[i].text())
            self.control_data["axis_stop"].append(self.stop_combos[i].currentText())
            self.control_data["auto_reference"].append(
                self.auto_ref_checkboxes[i].isChecked()
            )
            self.control_data["manual_function"].append(
                self.manual_move_combos[i].currentText()
            )

        super().accept()



from ophyd import Signal

import numpy as np

from epics import caput, caget


class EpicsFieldSignal(Signal):
    """A Signal used to address a single field of an EPICS PV."""

    def __init__(
        self,
        read_pv_name,
        name,
        value=0.0,
        timestamp=None,
        parent=None,
        labels=None,
        kind="hinted",
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        conversion_function=None,
        set_conversion_function=None,
        putFunc=None,
    ):
        """
        Parameters
        ----------
        read_pv_name : str
            The name of the used PV.field
        name : string, keyword only
        value : any, optional
            The initial value
        timestamp : float, optional
            The timestamp associated with the initial value. Defaults to
            the current local time.
        parent : Device, optional
            The parent Device holding this signal
        kind : a member the Kind IntEnum (or equivalent integer), optional
            Default is Kind.normal. See Kind for options.
        tolerance : any, optional
            The absolute tolerance associated with the value
        rtolerance : any, optional
            The relative tolerance associated with the value, used in
            set_and_wait as follows
            .. math::
              |setpoint - readback| \\leq (tolerance + rtolerance * |readback|)
        cl : namespace, optional
            Control Layer.  Must provide 'get_pv' and 'thread_class'
        attr_name : str, optional
            The parent Device attribute name that corresponds to this
            Signal
        conversion_function : function, default lambda x: x
            This function is used to convert the given value for `put`
            to the actual value that the Signal should be set to
        set_conversion_function : function, default lambda x: x
            This function is used to convert the given value from `get`
            to the actual value that the User wants to see
        putFunc : function, default lambda x: x
            A function to be called when the value is put (e.g. to
            connect to other Signals)
        """
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )
        self.read_pv_name = read_pv_name
        if conversion_function is None:
            conversion_function = lambda x: x
        self.conversion_function = conversion_function
        if set_conversion_function is None:
            set_conversion_function = lambda x: x
        self.set_conversion_function = set_conversion_function
        self.put_values = None
        if putFunc is None:
            putFunc = lambda x: x
        self.putFunc = putFunc

    def just_readback(self):
        """Only returns the currently stored readback-value."""
        return self._readback

    def get(self):
        """Reads the Signals value. If there is a `conversion_function`,
        the value is transformed by it.

        Parameters
        ----------

        Returns
        -------

        """
        if self.read_pv_name is not None:
            getval = caget(self.read_pv_name)
            if (
                self.put_values is not None
                and type(getval) in [int, float, np.float64]
                and np.abs(getval - self.put_values[1]) <= 1e-3 * getval
            ):
                self._readback = self.put_values[0]
            else:
                self._readback = self.conversion_function(getval)
        return super().get()

    def put(self, value, *, timestamp=None, force=False, metadata=None, **kwargs):
        """Puts the Signals value. If there is a `set_conversion_function`,
        the value is transformed by it. Also calls `self.putFunc`.

        Parameters
        ----------
        value :

        * :

        timestamp :
             (Default value = None)
        force :
             (Default value = False)
        metadata :
             (Default value = None)
        **kwargs :


        Returns
        -------

        """
        self.putFunc(value)
        val = self.set_conversion_function(value)
        if self.read_pv_name is not None:
            caput(self.read_pv_name, val, wait=True)
            self.put_values = (value, val)
        super().put(val, timestamp=timestamp, force=force, metadata=metadata, **kwargs)


class EpicsFieldSignalRO(EpicsFieldSignal):
    """The read-only implementation of EpicsFieldSignal. The only
    difference is that it raises an error, when one tries to put a value.

    Parameters
    ----------

    Returns
    -------

    """

    def put(self, value, *, timestamp=None, force=False, metadata=None, **kwargs):
        """

        Parameters
        ----------
        value :

        * :

        timestamp :
             (Default value = None)
        force :
             (Default value = False)
        metadata :
             (Default value = None)
        **kwargs :


        Returns
        -------

        """
        raise Exception("EpicsFieldSignalRO does not support putting a value!")



from nomad_camels.utility import variables_handling


class Measurement_Channel:
    """Class that represents one single channel. Most important
    attributes are the name, and the device.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, name="", output=False, device=None, metadata=None):
        self.name = name
        self.output = output
        self.device = device
        self.metadata = metadata or {}

    def get_bluesky_name(self):
        """ """
        return f'{self.device}.{self.name.split(".")[-1]}'

    def get_pv_name(self):
        """Returns the name of the corresponding EPICS PV."""
        name = f'{variables_handling.preset}:{self.device}:{self.name.split(".")[-1]}'
        return name

    def get_meta_str(self):
        meta_str = ""
        for i, (k, v) in enumerate(self.metadata.items()):
            if i > 0:
                meta_str += "\n"
            meta_str += f"{k}: {v}"
        return meta_str


def from_pv_name(pv_name):
    """Returns the corresponding Channel-name of an EPICS PV.

    Parameters
    ----------
    pv_name :


    Returns
    -------

    """
    split_name = pv_name.split(":")
    return f"{split_name[1]}_{split_name[2]}"



from PySide6.QtCore import Qt, QRect, QPoint, QSize, Signal
from PySide6.QtWidgets import (
    QWidget,
    QLabel,
    QHBoxLayout,
    QToolButton,
    QLayout,
)


class TagWidget(QWidget):
    removing = Signal(QWidget)

    def __init__(self, text, parent=None):
        super().__init__(parent)
        self.text = text

        # Tighter layout: less horizontal and vertical margins
        layout = QHBoxLayout(self)
        layout.setContentsMargins(4, 2, 4, 2)  # smaller margins
        layout.setSpacing(2)  # reduce spacing between label and 'X'

        self.label = QLabel(text, self)
        layout.addWidget(self.label)

        # Smaller close button to appear more integrated
        self.close_button = QToolButton(self)
        self.close_button.setText("X")
        self.close_button.setFixedSize(12, 12)  # smaller button size
        self.close_button.clicked.connect(self.remove_self)

        # Style the button to blend well with the tag background
        self.close_button.setStyleSheet(
            """
        QToolButton {
            border: none; 
            padding: 0;
            margin: 0;
            font-weight: bold;
            color: #333;
            background: transparent;
        }
        QToolButton:hover {
            color: red;
        }
        """
        )

        layout.addWidget(self.close_button)

        # Style the tag widget as a cohesive "pill"
        self.setStyleSheet(
            """
        QWidget {
            background: #D0E9FF;
            border: 1px solid #8BBCE4;
            border-radius: 4px;
        }
        """
        )

    def remove_self(self):
        self.setParent(None)
        self.removing.emit(self)
        self.deleteLater()


class FlowLayout(QLayout):
    # A simple flow layout to arrange tags nicely
    def __init__(self, parent=None, margin=0, spacing=-1):
        super().__init__(parent)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)
        self.setSpacing(spacing)
        self.itemList = []
        self.widgetList = []

    def get_all_tags(self):
        return [widget.text for widget in self.widgetList]

    def addWidget(self, widget):
        self.widgetList.append(widget)
        widget.removing.connect(self.remove_widget)
        super().addWidget(widget)

    def remove_widget(self, widget):
        if widget in self.widgetList:
            self.widgetList.remove(widget)

    def addItem(self, item):
        self.itemList.append(item)

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList[index]
        return None

    def takeAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList.pop(index)
        return None

    def expandingDirections(self):
        return Qt.Orientations(Qt.Orientation(0))

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        return self.doLayout(QRect(0, 0, width, 0), True)

    def setGeometry(self, rect):
        super().setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()
        for item in self.itemList:
            size = size.expandedTo(item.sizeHint())
        margins = self.contentsMargins()
        size += QSize(
            margins.left() + margins.right(), margins.top() + margins.bottom()
        )
        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0

        for item in self.itemList:
            wid = item.widget()
            spaceX = self.spacing()
            spaceY = self.spacing()
            nextX = x + wid.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + wid.sizeHint().width() + spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(QRect(QPoint(x, y), wid.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, wid.sizeHint().height())

        return y + lineHeight - rect.y()



import pytest
import time
from unittest.mock import MagicMock, patch
from pytest_mock import mocker
from PySide6.QtWidgets import QApplication, QMessageBox
from PySide6.QtGui import QCloseEvent

from nomad_camels.main_classes.manual_control import (
    Manual_Control,
    Manual_Control_Config,
)


@pytest.fixture(scope="module")
def app():
    """Fixture for creating a QApplication instance"""
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    return app


@pytest.fixture
def manual_control(app):
    return Manual_Control(title="Test Control")


@pytest.fixture
def manual_control_config(app):
    return Manual_Control_Config(title="Test Control Config")


def test_manual_control_init(manual_control):
    assert manual_control.windowTitle() == "Test Control - NOMAD CAMELS"
    assert manual_control.name == "Test Control"
    assert manual_control.device is None
    assert manual_control.ophyd_device is None
    assert manual_control.device_list == []


def test_manual_control_close_event(manual_control):
    with patch(
        "nomad_camels.utility.device_handling.close_devices"
    ) as mock_close_devices:
        event = QCloseEvent()
        manual_control.closeEvent(event)
        mock_close_devices.assert_called_with([])
        assert event.isAccepted()


def test_manual_control_propagate_exception(manual_control):
    with pytest.raises(Exception, match="Test Exception"):
        manual_control.propagate_exception(Exception("Test Exception"))


def test_manual_control_start_device(app, manual_control):
    mock_device = MagicMock()
    mock_device.get_necessary_devices.return_value = ["device1", "device2"]
    mock_device.custom_name = "Test_Device"
    mock_device1 = MagicMock()
    mock_device1.custom_name = "device1"
    mock_device2 = MagicMock()
    mock_device2.custom_name = "device2"

    with patch(
        "nomad_camels.utility.variables_handling.devices",
        {"Test_Device": mock_device, "device1": mock_device1, "device2": mock_device2},
    ):
        manual_control.start_device("Test_Device")
        time.sleep(0.5)
        app.processEvents()
        assert manual_control.device == mock_device
        assert "device1" in manual_control.device_list
        assert "device2" in manual_control.device_list
        assert "Test_Device" in manual_control.device_list


def test_manual_control_start_multiple_devices(manual_control):
    device_names = ["device1", "device2"]
    with patch(
        "nomad_camels.utility.device_handling.InstantiateDevicesThread"
    ) as mock_thread:
        manual_control.start_multiple_devices(device_names)
        mock_thread.assert_called_with(device_names, False)
        assert not manual_control.isEnabled()


def test_manual_control_device_ready(manual_control):
    manual_control.instantiate_devices_thread = MagicMock()
    manual_control.instantiate_devices_thread.devices = {
        "Test Device": "ophyd_device_mock"
    }

    manual_control.device = MagicMock()
    manual_control.device.custom_name = "Test Device"

    manual_control.device_ready()
    assert manual_control.ophyd_device == "ophyd_device_mock"
    assert manual_control.isEnabled()


def test_manual_control_config_init(manual_control_config):
    assert manual_control_config.windowTitle() == "Test Control Config - NOMAD CAMELS"
    assert manual_control_config.control_type == "Manual_Control"
    assert manual_control_config.lineEdit_name.text() == "Manual_Control"


def test_manual_control_config_accept(manual_control_config):
    manual_control_config.lineEdit_name.setText("New Control Name")
    manual_control_config.accept()
    assert manual_control_config.control_data["name"] == "New Control Name"
    assert manual_control_config.control_data["control_type"] == "Manual_Control"


def test_manual_control_config_close_event(manual_control_config):
    event = QCloseEvent()
    with patch("PySide6.QtWidgets.QMessageBox.question", return_value=QMessageBox.Yes):
        manual_control_config.closeEvent(event)
        assert event.isAccepted()

    event = QCloseEvent()
    with patch("PySide6.QtWidgets.QMessageBox.question", return_value=QMessageBox.No):
        manual_control_config.closeEvent(event)
        assert not event.isAccepted()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'gradient_descent_step.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QGridLayout,
    QLabel,
    QSizePolicy,
    QWidget,
)

from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box


class Ui_Grad_Desc(object):
    """ """

    def setupUi(self, Grad_Desc):
        """

        Parameters
        ----------
        Grad_Desc :


        Returns
        -------

        """
        if not Grad_Desc.objectName():
            Grad_Desc.setObjectName("Grad_Desc")
        Grad_Desc.resize(400, 300)
        self.gridLayout = QGridLayout(Grad_Desc)
        self.gridLayout.setObjectName("gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.label_12 = QLabel(Grad_Desc)
        self.label_12.setObjectName("label_12")

        self.gridLayout.addWidget(self.label_12, 8, 2, 1, 1)

        self.lineEdit_learning_rate = Variable_Box(Grad_Desc)
        self.lineEdit_learning_rate.setObjectName("lineEdit_learning_rate")

        self.gridLayout.addWidget(self.lineEdit_learning_rate, 7, 0, 1, 1)

        self.label_5 = QLabel(Grad_Desc)
        self.label_5.setObjectName("label_5")

        self.gridLayout.addWidget(self.label_5, 4, 1, 1, 1)

        self.label = QLabel(Grad_Desc)
        self.label.setObjectName("label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.label_9 = QLabel(Grad_Desc)
        self.label_9.setObjectName("label_9")

        self.gridLayout.addWidget(self.label_9, 6, 2, 1, 1)

        self.label_3 = QLabel(Grad_Desc)
        self.label_3.setObjectName("label_3")

        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 1)

        self.label_10 = QLabel(Grad_Desc)
        self.label_10.setObjectName("label_10")

        self.gridLayout.addWidget(self.label_10, 8, 0, 1, 1)

        self.label_11 = QLabel(Grad_Desc)
        self.label_11.setObjectName("label_11")

        self.gridLayout.addWidget(self.label_11, 8, 1, 1, 1)

        self.label_2 = QLabel(Grad_Desc)
        self.label_2.setObjectName("label_2")

        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)

        self.lineEdit_opt_func = Variable_Box(Grad_Desc)
        self.lineEdit_opt_func.setObjectName("lineEdit_opt_func")

        self.gridLayout.addWidget(self.lineEdit_opt_func, 2, 1, 1, 2)

        self.lineEdit_max_val = Variable_Box(Grad_Desc)
        self.lineEdit_max_val.setObjectName("lineEdit_max_val")

        self.gridLayout.addWidget(self.lineEdit_max_val, 5, 2, 1, 1)

        self.label_6 = QLabel(Grad_Desc)
        self.label_6.setObjectName("label_6")

        self.gridLayout.addWidget(self.label_6, 4, 2, 1, 1)

        self.checkBox_plot_steps = QCheckBox(Grad_Desc)
        self.checkBox_plot_steps.setObjectName("checkBox_plot_steps")

        self.gridLayout.addWidget(self.checkBox_plot_steps, 3, 0, 1, 3)

        self.lineEdit_min_val = Variable_Box(Grad_Desc)
        self.lineEdit_min_val.setObjectName("lineEdit_min_val")

        self.gridLayout.addWidget(self.lineEdit_min_val, 5, 1, 1, 1)

        self.lineEdit_starting_val = Variable_Box(Grad_Desc)
        self.lineEdit_starting_val.setObjectName("lineEdit_starting_val")

        self.gridLayout.addWidget(self.lineEdit_starting_val, 5, 0, 1, 1)

        self.lineEdit_momentum = Variable_Box(Grad_Desc)
        self.lineEdit_momentum.setObjectName("lineEdit_momentum")

        self.gridLayout.addWidget(self.lineEdit_momentum, 7, 2, 1, 1)

        self.lineEdit_largest_step = Variable_Box(Grad_Desc)
        self.lineEdit_largest_step.setObjectName("lineEdit_largest_step")

        self.gridLayout.addWidget(self.lineEdit_largest_step, 9, 1, 1, 1)

        self.label_8 = QLabel(Grad_Desc)
        self.label_8.setObjectName("label_8")

        self.gridLayout.addWidget(self.label_8, 6, 1, 1, 1)

        self.lineEdit_threshold = Variable_Box(Grad_Desc)
        self.lineEdit_threshold.setObjectName("lineEdit_threshold")

        self.gridLayout.addWidget(self.lineEdit_threshold, 7, 1, 1, 1)

        self.comboBox_output_channel = QComboBox(Grad_Desc)
        self.comboBox_output_channel.setObjectName("comboBox_output_channel")

        self.gridLayout.addWidget(self.comboBox_output_channel, 1, 1, 1, 2)

        self.lineEdit_smallest_step = Variable_Box(Grad_Desc)
        self.lineEdit_smallest_step.setObjectName("lineEdit_smallest_step")

        self.gridLayout.addWidget(self.lineEdit_smallest_step, 9, 0, 1, 1)

        self.comboBox_extremum_type = QComboBox(Grad_Desc)
        self.comboBox_extremum_type.setObjectName("comboBox_extremum_type")

        self.gridLayout.addWidget(self.comboBox_extremum_type, 0, 1, 1, 2)

        self.lineEdit_max_n_steps = Variable_Box(Grad_Desc)
        self.lineEdit_max_n_steps.setObjectName("lineEdit_max_n_steps")

        self.gridLayout.addWidget(self.lineEdit_max_n_steps, 9, 2, 1, 1)

        self.label_4 = QLabel(Grad_Desc)
        self.label_4.setObjectName("label_4")

        self.gridLayout.addWidget(self.label_4, 4, 0, 1, 1)

        self.label_7 = QLabel(Grad_Desc)
        self.label_7.setObjectName("label_7")

        self.gridLayout.addWidget(self.label_7, 6, 0, 1, 1)

        QWidget.setTabOrder(self.comboBox_extremum_type, self.comboBox_output_channel)
        QWidget.setTabOrder(self.comboBox_output_channel, self.lineEdit_opt_func)
        QWidget.setTabOrder(self.lineEdit_opt_func, self.checkBox_plot_steps)
        QWidget.setTabOrder(self.checkBox_plot_steps, self.lineEdit_starting_val)
        QWidget.setTabOrder(self.lineEdit_starting_val, self.lineEdit_min_val)
        QWidget.setTabOrder(self.lineEdit_min_val, self.lineEdit_max_val)
        QWidget.setTabOrder(self.lineEdit_max_val, self.lineEdit_learning_rate)
        QWidget.setTabOrder(self.lineEdit_learning_rate, self.lineEdit_threshold)
        QWidget.setTabOrder(self.lineEdit_threshold, self.lineEdit_momentum)
        QWidget.setTabOrder(self.lineEdit_momentum, self.lineEdit_smallest_step)
        QWidget.setTabOrder(self.lineEdit_smallest_step, self.lineEdit_largest_step)
        QWidget.setTabOrder(self.lineEdit_largest_step, self.lineEdit_max_n_steps)

        self.retranslateUi(Grad_Desc)

        QMetaObject.connectSlotsByName(Grad_Desc)

    # setupUi

    def retranslateUi(self, Grad_Desc):
        """

        Parameters
        ----------
        Grad_Desc :


        Returns
        -------

        """
        Grad_Desc.setWindowTitle(QCoreApplication.translate("Grad_Desc", "Form", None))
        self.label_12.setText(
            QCoreApplication.translate("Grad_Desc", "Max Number of Steps", None)
        )
        self.label_5.setText(
            QCoreApplication.translate("Grad_Desc", "Minimum Value", None)
        )
        self.label.setText(
            QCoreApplication.translate("Grad_Desc", "Extremum-Type:", None)
        )
        self.label_9.setText(QCoreApplication.translate("Grad_Desc", "Momentum", None))
        self.label_3.setText(
            QCoreApplication.translate("Grad_Desc", "Optimization Function:", None)
        )
        self.label_10.setText(
            QCoreApplication.translate("Grad_Desc", "Smallest Step", None)
        )
        self.label_11.setText(
            QCoreApplication.translate("Grad_Desc", "Largest Step", None)
        )
        self.label_2.setText(
            QCoreApplication.translate("Grad_Desc", "Output-Channel:", None)
        )
        self.label_6.setText(
            QCoreApplication.translate("Grad_Desc", "Maximum Value", None)
        )
        self.checkBox_plot_steps.setText(
            QCoreApplication.translate("Grad_Desc", "Plot Steps", None)
        )
        self.label_8.setText(QCoreApplication.translate("Grad_Desc", "Threshold", None))
        self.label_4.setText(
            QCoreApplication.translate("Grad_Desc", "Starting Value", None)
        )
        self.label_7.setText(
            QCoreApplication.translate("Grad_Desc", "Learning Rate", None)
        )

    # retranslateUi



from PySide6.QtWidgets import QLineEdit, QVBoxLayout, QWidget, QWidgetAction
from PySide6.QtCore import Qt

# from PySide6.QtGui import QStandardItem

from nomad_camels.utility import variables_handling


# class VariableItem(QStandardItem):
#     def __init__(self, value=None):
#         super().__init__(value)
#         self.setToolTip('test')


class Variable_Box(QLineEdit):
    """QLineEdit that checks its contents for validity and then color-
    codes the background.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, parent=None):
        super().__init__(parent)
        # self.setToolTip("test")
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.context_menu)
        self.textChanged.connect(self.check_string)
        self.check_string()

    def setEnabled(self, a0: bool) -> None:
        """

        Parameters
        ----------
        a0: bool :


        Returns
        -------

        """
        if a0:
            self.check_string()
        else:
            self.setStyleSheet("")
        super().setEnabled(a0)

    def check_string(self):
        """Check the string and then set the corresponding background-color."""
        if not self.text():
            self.setStyleSheet(
                f'background-color: rgb{variables_handling.get_color("white", True)}'
            )
        if variables_handling.check_eval(self.text()):
            self.setStyleSheet(
                f'background-color: rgb{variables_handling.get_color("green", True)}'
            )
        else:
            self.setStyleSheet(
                f'background-color: rgb{variables_handling.get_color("red", True)}'
            )

    def get_value(self):
        """ """
        return variables_handling.get_eval(self.text())

    def context_menu(self, pos):
        """Generates the right-click-menu.
        There are entries for inserting (replace) and appending the
        variables, channels, functions and operators.

        Parameters
        ----------
        pos :


        Returns
        -------

        """
        menu = self.createStandardContextMenu()
        search_bar = QLineEdit(menu)
        search_bar.setPlaceholderText("Search...")
        # Setting up layout for the menu
        layout = QVBoxLayout()
        layout.addWidget(search_bar)
        container = QWidget()
        container.setLayout(layout)

        # Adding the container widget to the menu
        action = QWidgetAction(menu)
        action.setDefaultWidget(container)
        menu.insertAction(menu.actions()[0], action)
        # putting the returned actions somewhere is necessary, otherwise
        # there will be none inside the single menus
        (self.channel_menu, self.variable_menu, self.function_menu), _ = (
            variables_handling.get_menus(self.insert_variable)
        )
        first_act = menu.actions()[1]
        menu.insertMenu(first_act, self.channel_menu)
        menu.insertMenu(first_act, self.variable_menu)
        menu.insertMenu(first_act, self.function_menu)
        # menu.insertMenu(first_act, operator_menu)
        menu.insertSeparator(first_act)
        (
            self.channel_menu2,
            self.variable_menu2,
            self.operator_menu2,
            self.function_menu2,
        ), __ = variables_handling.get_menus(self.append_variable, "Append")
        menu.insertMenu(first_act, self.channel_menu2)
        menu.insertMenu(first_act, self.variable_menu2)
        menu.insertMenu(first_act, self.function_menu2)
        menu.insertMenu(first_act, self.operator_menu2)
        menu.insertSeparator(first_act)

        search_bar.textChanged.connect(self.filter_actions)

        menu.exec_(self.mapToGlobal(pos))

    def filter_actions(self, query):
        for menu in [
            self.channel_menu,
            self.variable_menu,
            self.function_menu,
            self.channel_menu2,
            self.variable_menu2,
            self.function_menu2,
        ]:
            if not any(
                query.lower() in action.text().lower() for action in menu.actions()
            ):
                menu.setEnabled(False)
                continue
            elif not menu.isEnabled():
                menu.setEnabled(True)
            for action in menu.actions():
                action.setVisible(query.lower() in action.text().lower())

    def append_variable(self, variable):
        """Used for the single actions of the context menu.

        Parameters
        ----------
        variable :


        Returns
        -------

        """
        self.setText(self.text() + f"{variable}")

    def insert_variable(self, variable):
        """Used for the single actions of the context menu.

        Parameters
        ----------
        variable :


        Returns
        -------

        """
        self.setText(f"{variable}")
